#!/usr/bin/env python3
"""Contains the different formats of CPDs used in PGM"""
from __future__ import division

from itertools import product
from warnings import warn
import numbers

import numpy as np

from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.extern import tabulate
from pgmpy.extern import six
from pgmpy.extern.six.moves import range, zip
from pgmpy.utils import StateNameInit
from pgmpy.utils import StateNameDecorator


class TabularCPD(DiscreteFactor):
    """
    Defines the conditional probability distribution table (cpd table)

    Example
    -------
    For a distribution of P(grade|diff, intel)

    +-------+--------------------+------------------+
    |diff   |      easy          |    hard          |
    +-------+-----+------+-------+------+----+------+
    |intel  |dumb |  avg | smart | dumb |avg |smart |
    +-------+-----+------+-------+------+----+------+
    |gradeA |0.1  |  0.1 |  0.1  | 0.1  |0.1 | 0.1  |
    +-------+-----+------+-------+------+----+------+
    |gradeB |0.1  |  0.1 |  0.1  | 0.1  |0.1 | 0.1  |
    +-------+-----+------+-------+------+----+------+
    |gradeC |0.8  |  0.8 |  0.8  | 0.8  |0.8 | 0.8  |
    +-------+-----+------+-------+------+----+------+

    values should be
    [[0.1,0.1,0.1,0.1,0.1,0.1],
    [0.1,0.1,0.1,0.1,0.1,0.1],
    [0.8,0.8,0.8,0.8,0.8,0.8]]

    >>> cpd = TabularCPD('grade',3,[[0.1,0.1,0.1,0.1,0.1,0.1],
                                    [0.1,0.1,0.1,0.1,0.1,0.1],
                                    [0.8,0.8,0.8,0.8,0.8,0.8]],
                                    evidence=['diff', 'intel'], evidence_card=[2,3])
    >>> print(cpd)
    +---------+---------+---------+---------+---------+---------+---------+
    | diff    | diff_0  | diff_0  | diff_0  | diff_1  | diff_1  | diff_1  |
    +---------+---------+---------+---------+---------+---------+---------+
    | intel   | intel_0 | intel_1 | intel_2 | intel_0 | intel_1 | intel_2 |
    +---------+---------+---------+---------+---------+---------+---------+
    | grade_0 | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     |
    +---------+---------+---------+---------+---------+---------+---------+
    | grade_1 | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     |
    +---------+---------+---------+---------+---------+---------+---------+
    | grade_2 | 0.8     | 0.8     | 0.8     | 0.8     | 0.8     | 0.8     |
    +---------+---------+---------+---------+---------+---------+---------+
    >>> cpd.values
    array([[[ 0.1,  0.1,  0.1],
            [ 0.1,  0.1,  0.1]],

           [[ 0.1,  0.1,  0.1],
            [ 0.1,  0.1,  0.1]],

           [[ 0.8,  0.8,  0.8],
            [ 0.8,  0.8,  0.8]]])
    >>> cpd.variables
    ['grade', 'diff', 'intel']
    >>> cpd.cardinality
    array([3, 2, 3])
    >>> cpd.variable
    'grade'
    >>> cpd.variable_card
    3

    Parameters
    ----------
    variable: int, string (any hashable python object)
        The variable whose CPD is defined.

    variable_card: integer
        cardinality of variable

    values: 2d array, 2d list or 2d tuple
        values of the cpd table

    evidence: array-like
        evidences(if any) w.r.t. which cpd is defined

    evidence_card: integer, array-like
        cardinality of evidences (if any)

    Public Methods
    --------------
    get_values()
    marginalize([variables_list])
    normalize()
    reduce([values_list])
    """
    @StateNameInit()
    def __init__(self, variable, variable_card, values,
                 evidence=None, evidence_card=None):

        self.variable = variable
        self.variable_card = None

        variables = [variable]

        if not isinstance(variable_card, numbers.Integral):
            raise TypeError("Event cardinality must be an integer")
        self.variable_card = variable_card

        cardinality = [variable_card]
        if evidence_card is not None:
            if isinstance(evidence_card, numbers.Real):
                raise TypeError("Evidence card must be a list of numbers")
            cardinality.extend(evidence_card)

        if evidence is not None:
            if isinstance(evidence, six.string_types):
                raise TypeError("Evidence must be list, tuple or array of strings.")
            variables.extend(evidence)
            if not len(evidence_card) == len(evidence):
                raise ValueError("Length of evidence_card doesn't match length of evidence")

        values = np.array(values)
        if values.ndim != 2:
            raise TypeError("Values must be a 2D list/array")

        super(TabularCPD, self).__init__(variables, cardinality, values.flatten('C'),
                                         state_names=self.state_names)

    def __repr__(self):
        var_str = '<TabularCPD representing P({var}:{card}'.format(
            var=self.variable, card=self.variable_card)

        evidence = self.variables[1:]
        evidence_card = self.cardinality[1:]
        if evidence:
            evidence_str = ' | ' + ', '.join(['{var}:{card}'.format(var=var, card=card)
                                              for var, card in zip(evidence, evidence_card)])
        else:
            evidence_str = ''

        return var_str + evidence_str + ') at {address}>'.format(address=hex(id(self)))

    def get_values(self):
        """
        Returns the cpd
        
        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd = TabularCPD('grade', 3, [[0.1, 0.1],
        ...                               [0.1, 0.1],
        ...                               [0.8, 0.8]],
        ...                  evidence='evi1', evidence_card=2)
        >>> cpd.get_values()
        array([[ 0.1,  0.1],
               [ 0.1,  0.1],
               [ 0.8,  0.8]])

        """
        if self.variable in self.variables:
            return self.values.reshape(self.cardinality[0], np.prod(self.cardinality[1:]))
        else:
            return self.values.reshape(1, np.prod(self.cardinality))

    def __str__(self):
        if six.PY2:
            return self._make_table_str("grid")
        else:
            return self._make_table_str("fancy_grid")

    def _str(self, phi_or_p="p", tablefmt="fancy_grid"):
        return super(self, TabularCPD)._str(phi_or_p, tablefmt)

    def _make_table_str(self, tablefmt="fancy_grid", print_state_names=True):
        headers_list = []
        # build column headers

        evidence = self.variables[1:]
        evidence_card = self.cardinality[1:]
        if evidence:
            col_indexes = np.array(list(product(*[range(i) for i in evidence_card])))
            if self.state_names and print_state_names:
                for i in range(len(evidence_card)):
                    column_header = [evidence[i]] + ['{var}({state})'.format
                                                     (var=evidence[i],
                                                      state=self.state_names[evidence[i]][d])
                                                     for d in col_indexes.T[i]]
                    headers_list.append(column_header)
            else:
                for i in range(len(evidence_card)):
                    column_header = [evidence[i]] + ['{s}_{d}'.format(s=evidence[i], d=d) for d in col_indexes.T[i]]
                    headers_list.append(column_header)

        # Build row headers
        if self.state_names and print_state_names:
            variable_array = [['{var}({state})'.format
                               (var=self.variable, state=self.state_names[self.variable][i])
                               for i in range(self.variable_card)]]
        else:
            variable_array = [['{s}_{d}'.format(s=self.variable, d=i) for i in range(self.variable_card)]]
        # Stack with data
        labeled_rows = np.hstack((np.array(variable_array).T, self.get_values())).tolist()
        # No support for multi-headers in tabulate
        cdf_str = tabulate(headers_list + labeled_rows, tablefmt=tablefmt)
        return cdf_str

    def copy(self):
        """
        Returns a copy of the TabularCPD object.

        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd = TabularCPD('grade', 2,
        ...                  [[0.7, 0.6, 0.6, 0.2],[0.3, 0.4, 0.4, 0.8]],
        ...                  ['intel', 'diff'], [2, 2])
        >>> copy = cpd.copy()
        >>> copy.variable
        'grade'
        >>> copy.variable_card
        2
        >>> copy.evidence
        ['intel', 'diff']
        >>> copy.values
        array([[[ 0.7,  0.6],
                [ 0.6,  0.2]],

               [[ 0.3,  0.4],
                [ 0.4,  0.8]]])
        """
        evidence = self.variables[1:] if len(self.variables) > 1 else None
        evidence_card = self.cardinality[1:] if len(self.variables) > 1 else None
        return TabularCPD(self.variable, self.variable_card, self.get_values(),
                          evidence, evidence_card)

    def normalize(self, inplace=True):
        """
        Normalizes the cpd table.

        Parameters
        ----------
        inplace: boolean
            If inplace=True it will modify the CPD itself, else would return
            a new CPD

        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd_table = TabularCPD('grade', 2,
        ...                        [[0.7, 0.2, 0.6, 0.2],[0.4, 0.4, 0.4, 0.8]],
        ...                        ['intel', 'diff'], [2, 2])
        >>> cpd_table.normalize()
        >>> cpd_table.get_values()
        array([[ 0.63636364,  0.33333333,  0.6       ,  0.2       ],
               [ 0.36363636,  0.66666667,  0.4       ,  0.8       ]])
        """
        tabular_cpd = self if inplace else self.copy()
        cpd = tabular_cpd.get_values()
        tabular_cpd.values = (cpd / cpd.sum(axis=0)).reshape(tabular_cpd.cardinality)
        if not inplace:
            return tabular_cpd

    def marginalize(self, variables, inplace=True):
        """
        Modifies the cpd table with marginalized values.

        Parameters
        ---------
        variables: list, array-like
            list of variable to be marginalized

        inplace: boolean
            If inplace=True it will modify the CPD itself, else would return
            a new CPD

        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd_table = TabularCPD('grade', 2,
        ...                        [[0.7, 0.6, 0.6, 0.2],[0.3, 0.4, 0.4, 0.8]],
        ...                        ['intel', 'diff'], [2, 2])
        >>> cpd_table.marginalize(['diff'])
        >>> cpd_table.get_values()
        array([[ 0.65,  0.4 ],
                [ 0.35,  0.6 ]])
        """
        if self.variable in variables:
            raise ValueError("Marginalization not allowed on the variable on which CPD is defined")

        tabular_cpd = self if inplace else self.copy()

        super(TabularCPD, tabular_cpd).marginalize(variables)
        tabular_cpd.normalize()

        if not inplace:
            return tabular_cpd

    @StateNameDecorator(argument='values', return_val=None)
    def reduce(self, values, inplace=True):
        """
        Reduces the cpd table to the context of given variable values.

        Parameters
        ----------
        values: list, array-like
            A list of tuples of the form (variable_name, variable_state).

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new factor.

        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd_table = TabularCPD('grade', 2,
        ...                        [[0.7, 0.6, 0.6, 0.2],[0.3, 0.4, 0.4, 0.8]],
        ...                        ['intel', 'diff'], [2, 2])
        >>> cpd_table.reduce([('diff', 0)])
        >>> cpd_table.get_values()
        array([[ 0.7,  0.6],
               [ 0.3,  0.4]])
        """
        if self.variable in (value[0] for value in values):
            raise ValueError("Reduce not allowed on the variable on which CPD is defined")

        tabular_cpd = self if inplace else self.copy()

        super(TabularCPD, tabular_cpd).reduce(values)
        tabular_cpd.normalize()

        if not inplace:
            return tabular_cpd

    def to_factor(self):
        """
        Returns an equivalent factor with the same variables, cardinality, values as that of the cpd

        Examples
        --------
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> cpd = TabularCPD('grade', 3, [[0.1, 0.1],
        ...                               [0.1, 0.1],
        ...                               [0.8, 0.8]],
        ...                  evidence='evi1', evidence_card=2)
        >>> factor = cpd.to_factor()
        >>> factor
        <DiscreteFactor representing phi(grade:3, evi1:2) at 0x7f847a4f2d68>
        """
        return DiscreteFactor(self.variables, self.cardinality, self.values)

    def reorder_parents(self, new_order, inplace=True):
        """
        Returns a new cpd table according to provided order

        Parameters
        ----------
        new_order: list
            list of new ordering of variables

        inplace: boolean
            If inplace == True it will modify the CPD itself
            otherwise new value will be returned without affecting old values

        Examples
        --------
        Consider a CPD P(grade| diff, intel)
        >>> cpd = TabularCPD('grade',3,[[0.1,0.1,0.1,0.1,0.1,0.1],
                                        [0.1,0.1,0.1,0.1,0.1,0.1],
                                        [0.8,0.8,0.8,0.8,0.8,0.8]],
                                    evidence=['diff', 'intel'], evidence_card=[2,3])
        >>> print(cpd)
        +---------+---------+---------+---------+---------+---------+---------+
        | diff    | diff_0  | diff_0  | diff_0  | diff_1  | diff_1  | diff_1  |
        +---------+---------+---------+---------+---------+---------+---------+
        | intel   | intel_0 | intel_1 | intel_2 | intel_0 | intel_1 | intel_2 |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_0 | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_1 | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_2 | 0.8     | 0.8     | 0.8     | 0.8     | 0.8     | 0.8     |
        +---------+---------+---------+---------+---------+---------+---------+
        >>> cpd.values
        array([[[ 0.1,  0.1,  0.1],
                [ 0.1,  0.1,  0.1]],

               [[ 0.1,  0.1,  0.1],
                [ 0.1,  0.1,  0.1]],

               [[ 0.8,  0.8,  0.8],
                [ 0.8,  0.8,  0.8]]])
        >>> cpd.variables
        ['grade', 'diff', 'intel']
        >>> cpd.cardinality
        array([3, 2, 3])
        >>> cpd.variable
        'grade'
        >>> cpd.variable_card
        3

        >>> cpd.reorder_parents(['intel', 'diff'])
        array([[ 0.1,  0.1,  0.2,  0.2,  0.1,  0.1],
               [ 0.1,  0.1,  0.1,  0.1,  0.1,  0.1],
               [ 0.8,  0.8,  0.7,  0.7,  0.8,  0.8]])
        >>> print(cpd)
        +---------+---------+---------+---------+---------+---------+---------+
        | intel   | intel_0 | intel_0 | intel_1 | intel_1 | intel_2 | intel_2 |
        +---------+---------+---------+---------+---------+---------+---------+
        | diff    | diff_0  | diff_1  | diff_0  | diff_1  | diff_0  | diff_1  |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_0 | 0.1     | 0.1     | 0.2     | 0.2     | 0.1     | 0.1     |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_1 | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     | 0.1     |
        +---------+---------+---------+---------+---------+---------+---------+
        | grade_2 | 0.8     | 0.8     | 0.7     | 0.7     | 0.8     | 0.8     |
        +---------+---------+---------+---------+---------+---------+---------+

        >>> cpd.values
        array([[[ 0.1,  0.1],
                [ 0.2,  0.2],
                [ 0.1,  0.1]],

               [[ 0.1,  0.1],
                [ 0.1,  0.1],
                [ 0.1,  0.1]],

               [[ 0.8,  0.8],
                [ 0.7,  0.7],
                [ 0.8,  0.8]]])

        >>> cpd.variables
        ['grade', 'intel', 'diff']
        >>> cpd.cardinality
        array([3, 3, 2])
        >>> cpd.variable
        'grade'
        >>> cpd.variable_card
        3
        """
        if (len(self.variables) <= 1 or (set(new_order) - set(self.variables)) or
                (set(self.variables[1:]) - set(new_order))):
            raise ValueError("New order either has missing or extra arguments")
        else:
            if new_order != self.variables[1:]:
                evidence = self.variables[1:]
                evidence_card = self.cardinality[1:]
                card_map = dict(zip(evidence, evidence_card))
                old_pos_map = dict(zip(evidence, range(len(evidence))))
                trans_ord = [0] + [(old_pos_map[letter] + 1) for letter in new_order]
                new_values = np.transpose(self.values, trans_ord)

                if inplace:
                    variables = [self.variables[0]] + new_order
                    cardinality = [self.variable_card] + [card_map[var] for var in new_order]
                    super(TabularCPD, self).__init__(variables, cardinality, new_values.flatten('C'))
                    return self.get_values()
                else:
                    return new_values.reshape(self.cardinality[0], np.prod([card_map[var] for var in new_order]))
            else:
                warn("Same ordering provided as current")
                return self.get_values()

    def get_evidence(self):
        return self.variables[:0:-1]

# Commenting out because not used anywhere for now and not implemented in a very good way.
# class TreeCPD(nx.DiGraph):
#     """
#     Base Class for Tree CPD.
#     """
#     def __init__(self, data=None):
#         """
#         Base Class for Tree CPD.
#
#         Parameters
#         ----------
#         data: input tree
#             Data to initialize the tree. If data=None (default) an empty
#             tree is created. The data can be an edge list with label for
#             each edge. Label should be the observed value of the variable.
#
#         Examples
#         --------
#         For P(A|B, C, D), to construct a tree like:
#
#                     B
#              0 /        \1
#               /          \
#         P(A|b_0)          C
#                    0/         \1
#                    /           \
#             P(A|b_1, c_0)      D
#                           0/       \
#                           /         \
#             P(A|b_1,c_1,d_0)      P(A|b_1,c_1,d_1)
#
#         >>> from pgmpy.factors import TreeCPD, DiscreteFactor
#         >>> tree = TreeCPD([('B', DiscreteFactor(['A'], [2], [0.8, 0.2]), '0'),
#         ...                 ('B', 'C', '1'),
#         ...                 ('C', DiscreteFactor(['A'], [2], [0.1, 0.9]), '0'),
#         ...                 ('C', 'D', '1'),
#         ...                 ('D', DiscreteFactor(['A'], [2], [0.9, 0.1]), '0'),
#         ...                 ('D', DiscreteFactor(['A'], [2], [0.4, 0.6]), '1')])
#         """
#         nx.DiGraph.__init__(self)
#         # TODO: Check cycles and self loops.
#         if data:
#             for edge in data:
#                 if len(edge) != 3:
#                     raise ValueError("Each edge tuple must have 3 values (u, v, label).")
#                 self.add_edge(edge[0], edge[1], label=edge[2])
#
#     def add_edge(self, u, v, label):
#         """
#         Add an edge between u and v.
#
#         The nodes u and v will be automatically added if they are
#         not already in the graph.
#
#         Parameters
#         ----------
#         u,v: nodes
#             Nodes can be any hashable (and not None) Python object.
#         label: string
#             Label should be value of the variable observed.
#             (underscore separated if multiple variables)
#         attr_dict: dictionary, optional (default= no attributes)
#             Dictionary of edge attributes. Key/Value pairs will
#             update existing data associated with the edge.
#         attr: Keyword arguments, optional
#             Edge data can be assigned using keyword arguments.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import TreeCPD, DiscreteFactor
#         >>> tree = TreeCPD([('B', DiscreteFactor(['A'], [2], [0.8, 0.2]), 0),
#         ...                 ('B', 'C', 1)])
#         >>> tree.add_edge('C', DiscreteFactor(['A'], [2], [0.1, 0.9]), label=0)
#         """
#         if u != v:
#             if u in self.nodes() and v in self.nodes() and nx.has_path(self, v, u):
#                 # check if adding edge (u, v) forms a cycle
#                 raise ValueError(
#                     'Loops are not allowed. Adding the edge from (%s->%s) forms a loop.' % (u, v))
#             else:
#                 super(TreeCPD, self).add_edge(u, v, label=label)
#         else:
#             raise ValueError('Self loops are not allowed. Edge (%s->%s) forms a self loop.' % (u, v))
#
#     def add_edges_from(self, ebunch):
#         """
#         Add all the edges in ebunch.
#
#         Parameters
#         ----------
#         ebunch : container of edges
#             Each edge given in the container will be added to the
#             graph. The edges must be given as as 3-tuples (u,v,label).
#         attr_dict : dictionary, optional (default= no attributes)
#             Dictionary of edge attributes.  Key/value pairs will
#             update existing data associated with each edge.
#         attr : keyword arguments, optional
#             Edge data (or labels or objects) can be assigned using
#             keyword arguments.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import TreeCPD, DiscreteFactor
#         >>> tree = TreeCPD()
#         >>> tree.add_edges_from([('B', 'C', 1), ('C', 'D', 1),
#         ...                      ('D', DiscreteFactor(['A'], [2], [0.6, 0.4]))])
#         """
#         for edge in ebunch:
#             if len(edge) == 2:
#                 raise ValueError("Each edge tuple must have 3 values (u, v, label).")
#         nx.DiGraph.add_edges_from(self, [(edge[0], edge[1], {'label': edge[2]}) for edge in ebunch])
#
#     def to_tabular_cpd(self, parents_order=None):
#         edge_attributes = nx.get_edge_attributes(self, 'label')
#         edge_values = {}
#         edge_dict = {}
#         adjlist = {}
#         node_list = []
#         stack = []
#         values = []
#         cardinality = []
#
#         for edge in edge_attributes:
#             edge_dict.setdefault(edge[0], []).append(edge_attributes[edge])
#             if isinstance(edge[1], DiscreteFactor):
#                 variable = edge[1].scope()
#                 variable_card = edge[1].cardinality
#                 edge_values[(edge[0], edge[0] + edge_attributes.get(edge))] = edge[1].values.tolist()
#             else:
#                 edge_values[(edge[0], edge[0] + edge_attributes.get(edge))] = edge[1]
#         #adjlist
#         for source in self.nodes():
#             if not isinstance(source, DiscreteFactor):
#                 adjlist[source] = [i[1] for i in edge_attributes if not isinstance(i[1], DiscreteFactor) and i[0] == source]
#                 adjlist[source] = sorted(adjlist[source], key=lambda x: (len(nx.descendants(self, x)), x))
#
#         root = [node for node, in_degree in self.in_degree().items() if in_degree == 0][0]
#         stack.append(root)
#
#         #dfs
#         while stack:
#             top_node = stack[-1]
#             node_list.append(top_node)
#             stack = stack[:-1]
#             for end_node in adjlist[top_node]:
#                 stack.append(end_node)
#
#         for node in node_list:
#             cardinality.append(len(edge_dict[node]))
#
#         for i in product(*[range(index) for index in cardinality]):
#             edge_list = [a + str(b) for a, b in zip(node_list, i)]
#             current_node = root
#             for edge in edge_list:
#                 if (current_node, edge) in edge_values.keys():
#                     if not isinstance(edge_values[(current_node, edge)], list):
#                         current_node = edge_values[(current_node, edge)]
#                     else:
#                         values.append(edge_values[(current_node, edge)])
#                         break
#
#         values = np.array(values).flatten('F').reshape((len(values[0]), len(values)))
#         return TabularCPD(variable[0], int(variable_card[0]), values, node_list[::-1], cardinality)
#
#     def to_rule_cpd(self):
#         """
#         Returns a RuleCPD object which represents the TreeCPD
#
#         Examples
#         --------
#         >>> from pgmpy.factors import TreeCPD, DiscreteFactor
#         >>> tree = TreeCPD([('B', factors(['A'], [2], [0.8, 0.2]), '0'),
#         ...                 ('B', 'C', '1'),
#         ...                 ('C', factors(['A'], [2], [0.1, 0.9]), '0'),
#         ...                 ('C', 'D', '1'),
#         ...                 ('D', factors(['A'], [2], [0.9, 0.1]), '0'),
#         ...                 ('D', factors(['A'], [2], [0.4, 0.6]), '1')])
#         >>> tree.to_rule_cpd()
#
#         """
#         # TODO: This method assumes that factors class has a get_variable method. Check this after merging navin's PR.
#         root = [node for node, in_degree in self.in_degree().items() if in_degree == 0][0]
#         paths_root_to_factors = {target: path for target, path in nx.single_source_shortest_path(self, root).items() if
#                                  isinstance(target, DiscreteFactor)}
#         for node in self.nodes_iter():
#             if isinstance(node, DiscreteFactor):
#                 rule_cpd = RuleCPD(node.scope()[0])
#
#         for factor, path in paths_root_to_factors.items():
#             rule_key = []
#             for node_index in range(len(path) - 1):
#                 rule_key.append(path[node_index] + '_' + self.edge[path[node_index]][path[node_index + 1]]['label'])
#             for value_index in range(len(factor.values)):
#                 rule_key.append(factor.get_variables()[0] + '_' + str(value_index))
#                 rule_cpd.add_rules({tuple(sorted(rule_key)): factor.values[value_index]})
#         return rule_cpd
#
#
# class RuleCPD:
#     def __init__(self, variable, rules=None):
#         """
#         Base class for Rule CPD.
#
#         Parameters
#         ----------
#         variable: str
#             The variable for which the CPD is to be defined.
#
#         rules: dict. (optional)
#             dict of rules. Each rule should be in the form of
#             tuple_of_assignment: probability.
#             For example: ('A_0', 'J_0'): 0.8
#
#         Examples
#         --------
#         For constructing a RuleCPD on variable A with the following rules:
#             p1: <A_0, B_0; 0.8>
#             p2: <A_1, B_0; 0.2>
#             p3: <A_0, B_1, C_0; 0.4>
#             p4: <A_1, B_1, C_0; 0.6>
#             p5: <A_0, B_1, C_1; 0.9>
#             p6: <A_1, B_1, C_1; 0.1>
#
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD('A', {('A_0', 'B_0'): 0.8,
#         ...                      ('A_1', 'B_0'): 0.2,
#         ...                      ('A_0', 'B_1', 'C_0'): 0.4,
#         ...                      ('A_1', 'B_1', 'C_0'): 0.6,
#         ...                      ('A_0', 'B_1', 'C_1'): 0.9,
#         ...                      ('A_1', 'B_1', 'C_1'): 0.1})
#         """
#         self.variable = variable
#         if rules:
#             self.rules = {}
#             for rule, value in rules.items():
#                 self.rules[tuple(sorted(rule))] = value
#         else:
#             self.rules = {}
#         verify = self._verify()
#         if not verify[0]:
#             del self
#             raise ValueError(str(verify[1]) + " and " + str(verify[2]) + " point to the same assignment")
#
#     def _verify(self):
#         """
#         Verifies the RuleCPD for multiple values of the
#         assignment.
#         """
#         from itertools import combinations
#         for rule, another_rule in combinations(self.rules, 2):
#             rule, another_rule = (rule, another_rule) if len(rule) < len(another_rule) else (another_rule, rule)
#             if not set(rule) - set(another_rule):
#                 return False, rule, another_rule
#         return True,
#
#     def add_rules(self, rules):
#         """
#         Add one or more rules to the Rule CPD.
#
#         Parameters
#         ----------
#         rules: dict
#             dict of rules. Each rule should be in the form of
#             tuple_of_assignment: probability.
#             For example: ('A_0', 'J_0'): 0.8
#
#         Examples
#         --------
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD(variable='A')
#         >>> rule.add_rules({('A_0', 'B_0'): 0.8,
#         ...                 ('A_1', 'B_0'): 0.2})
#         """
#         for rule in rules:
#             self.rules[rule] = rules[rule]
#         verify = self._verify()
#         if not verify[0]:
#             for rule in rules:
#                 del(self.rules[rule])
#             raise ValueError(str(verify[1]) + " and " + str(verify[2]) + " point to the same assignment")
#
#     def scope(self):
#         """
#         Returns a set of variables which is the scope of the Rule CPD.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD('A', {('A_0', 'B_0'): 0.8,
#         >>>                      ('A_1', 'B_0'): 0.2,
#         >>>                      ('A_0', 'B_1', 'C_0'): 0.4,
#         >>>                      ('A_1', 'B_1', 'C_0'): 0.6,
#         >>>                      ('A_0', 'B_1', 'C_!'): 0.9,
#         >>>                      ('A_1', 'B_1', 'C_1'): 0.1}
#         >>> rule.scope()
#         {'A', 'B', 'C'}
#         """
#         scope = set()
#         for rule in self.rules:
#             scope.update([assignment.split('_')[0] for assignment in rule])
#         return scope
#
#     def cardinality(self, variable=None):
#         """
#         Returns a dict of variable: cardinality.
#
#         Parameters
#         ----------
#         variable: string, list
#             variable or list of variables whose cardinality will be returned.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD('A', {('A_0', 'B_0'): 0.8,
#         >>>                      ('A_1', 'B_0'): 0.2,
#         >>>                      ('A_0', 'B_1', 'C_0'): 0.4,
#         >>>                      ('A_1', 'B_1', 'C_0'): 0.6,
#         >>>                      ('A_0', 'B_1', 'C_!'): 0.9,
#         >>>                      ('A_1', 'B_1', 'C_1'): 0.1}
#         >>> rule.cardinality()
#         {'A': 2, 'B': 2, 'C': 2}
#         """
#         from itertools import chain
#         from collections import Counter
#         assignments = set(chain.from_iterable(self.rules))
#         cardinality = dict(Counter([element.split('_')[0] for element in assignments]))
#         if variable:
#             return cardinality[variable] if isinstance(variable, str) else {var: cardinality[var] for var in variable}
#         else:
#             return cardinality
#
#     def to_tabular_cpd(self, parents_order=None):
#         """
#         Returns an equivalent TabularCPD.
#
#         Parameters
#         ----------
#         parents_order: array-like. list, tuple. (optional)
#             The order of the evidence variables.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD('A', {('A_0', 'B_0'): 0.8,
#         >>>                      ('A_1', 'B_0'): 0.2,
#         >>>                      ('A_0', 'B_1', 'C_1'): 0.9,
#         >>>                      ('A_1', 'B_1', 'C_1'): 0.1,
#         >>>                      ('A_0', 'B_1', 'C_0', 'D_0'): 0.4,
#         >>>                      ('A_1', 'B_1', 'C_0', 'D_0'): 0.6,
#         >>>                      ('A_0', 'B_1', 'C_0', 'D_1'): 0.3,
#         >>>                      ('A_1', 'B_1', 'C_0', 'D_1'): 0.7})
#         >>> rule.to_tabular_cpd()
#         """
#         if not parents_order:
#             parents_order = sorted(self.scope() - {self.variable})
#         cardinality_dict = self.cardinality()
#         cardinality_product = np.product(list(cardinality_dict.values()))
#         tabular_cpd = [[0] * cardinality_product
#                        for _ in range(cardinality_dict[self.variable])]
#         for rule, value in self.rules:
#             start, end = 0, cardinality_product
#             for var in sorted(rule):
#                 if var.split('_')[0] != self.variable:
#                     start, end = (start + (end-start)/cardinality_dict[var] * int(var.split('_')[1]),
#                                   start + (end-start)/cardinality_dict[var] * (int(var.split('_')[1]) + 1))
#                 else:
#                     var_assignment = int(var.split('_')[1])
#             for index in range(start, end):
#                 tabular_cpd[var_assignment][index] = value
#
#         return TabularCPD(self.variable, cardinality_dict[self.variable], tabular_cpd,
#                           parents_order, [cardinality_dict[var] for var in parents_order])
#
#     def _merge(self):
#         """
#         Removes the variable from the rules and then merges the rules
#         having the same variables.
#         For example:
#         If we are given these rules:
#         ('A_0', 'B_0'): 0.8,
#         ('A_1', 'B_0'): 0.2,
#         ('A_0', 'B_1', 'C_1'): 0.9,
#         ('A_1', 'B_1', 'C_1'): 0.1,
#         ('A_0', 'B_1', 'C_0', 'D_0'): 0.4,
#         ('A_1', 'B_1', 'C_0', 'D_0'): 0.6,
#         ('A_0', 'B_1', 'C_0', 'D_1'): 0.3,
#         ('A_1', 'B_1', 'C_0', 'D_1'): 0.7
#
#         then after merging _merge will return this dict:
#         {('B_0',): array([ 0.8,  0.2]),
#          ('B_1', 'C_0', 'D_1'): array([ 0.3,  0.7]),
#          ('B_1', 'C_1'): array([ 0.9,  0.1]),
#          ('B_1', 'C_0', 'D_0'): array([ 0.4,  0.6])}
#         """
#         var_card = self.cardinality(self.variable)
#         dict_without_var = {}
#         for assignments in self.rules.keys():
#             dict_without_var[tuple(sorted([var for var in assignments if not var.startswith(self.variable)]))] = None
#         for key in dict_without_var:
#             value_list = []
#             for assign in range(var_card):
#                 value_list.append(self.rules[tuple(sorted(list(key) + [(self.variable + '_' + str(assign))]))])
#             dict_without_var[key] = np.array(value_list)
#         return dict_without_var
#
#     def to_tree_cpd(self):
#         """
#         Return a TreeCPD object which represents the RuleCPD.
#
#         Examples
#         --------
#         >>> from pgmpy.factors import RuleCPD
#         >>> rule = RuleCPD('A', {('A_0', 'B_0'): 0.8,
#         ...                      ('A_1', 'B_0'): 0.2,
#         ...                      ('A_0', 'B_1', 'C_1'): 0.9,
#         ...                      ('A_1', 'B_1', 'C_1'): 0.1,
#         ...                      ('A_0', 'B_1', 'C_0', 'D_0'): 0.4,
#         ...                      ('A_1', 'B_1', 'C_0', 'D_0'): 0.6,
#         ...                      ('A_0', 'B_1', 'C_0', 'D_1'): 0.3,
#         ...                      ('A_1', 'B_1', 'C_0', 'D_1'): 0.7})
#         >>> rule.to_tree_cpd()
#         <CPD.TreeCPD object at 0x7f6b6f952fd0>
#         """
#         from collections import OrderedDict
#         tree_cpd = TreeCPD()
#         merged_rules = OrderedDict(sorted(self._merge().items(), key=lambda t: len(t[0])))
#
#         for assignments, value in merged_rules.items():
#             for assignment_index in range(len(assignments) - 1):
#                 tree_cpd.add_edge(assignments[assignment_index].split('_')[0],
#                                   assignments[assignment_index+1].split('_')[0],
#                                   assignments[assignment_index].split('_')[1])
#             tree_cpd.add_edge(assignments[-1].split('_')[0],
#                               DiscreteFactor([self.variable], [len(value)], value),
#                               assignments[-1].split('_')[1])
#         return tree_cpd
#
#     def __str__(self):
#         from collections import OrderedDict
#         string = ""
#         for index, key in enumerate(OrderedDict(sorted(self.rules.items(), key=lambda t: len(t[0])))):
#             key_string = ', '.join(key)
#             string += 'p' + str(index) + ': <' + key_string + '; ' + str(self.rules[key]) + '>' + '\n'
#         return string



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import unittest
import warnings
import json

import numpy as np
import numpy.testing as np_test

from pgmpy.readwrite import ProbModelXMLReader, ProbModelXMLWriter, get_probmodel_data
from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD
from pgmpy.extern.six.moves import range
from pgmpy.extern import six

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        try:
            import xml.etree.ElementTree as etree
            print("running with ElementTree on Python 2.5+")
        except ImportError:
            warnings.warn("Failed to import ElementTree from any known place")


class TestProbModelXMLReaderString(unittest.TestCase):
    def setUp(self):
        string = """<ProbModelXML formatVersion="1.0">
      <ProbNet type="BayesianNetwork">
        <AdditionalConstraints>
            <Constraint name="MaxNumParents">
                <Argument name="numParents" value="5" />
            </Constraint>
        </AdditionalConstraints>
        <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
        <Language>English</Language>
        <AdditionalProperties>
            <Property name="elvira.title" value="X ray result"/>
        </AdditionalProperties>
        <DecisionCriteria>
            <Criterion name="cost">
                <AdditionalProperties/>
            </Criterion>
            <Criterion name="effectiveness">
                <AdditionalProperties/>
            </Criterion>
        </DecisionCriteria>
        <Variables>
            <Variable name="intelligence" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="smart"><AdditionalProperties /></State>
                    <State name="dumb"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="difficulty" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="difficult"><AdditionalProperties /></State>
                    <State name="easy"><AdditionalProperties /></State>
                </States>
            </Variable>
        </Variables>
        <Links>
            <Link directed="1">
                <Variable name="difficulty" />
                <Variable name="grade" />
                <Comment>Directed Edge from difficulty to grade</Comment>
                <Label>diff_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="intelligence" />
                <Variable name="grade" />
                <Comment>Directed Edge from intelligence to grade</Comment>
                <Label>intel_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="intelligence" />
                <Variable name="SAT" />
                <Comment>Directed Edge from intelligence to SAT</Comment>
                <Label>intel_to_sat</Label>
                <AdditionalProperties />
            </Link>
            <Link directed="1">
                <Variable name="grade" />
                <Variable name="recommendation_letter" />
                <Comment>Directed Edge from grade to recommendation_letter</Comment>
                <Label>grad_to_reco</Label>
                <AdditionalProperties />
            </Link>
        </Links>
        <Potentials>
          <Potential type="Tree/ADD" role="Utility">
                <UtilityVariable name="U1" />
                <Variables>
                    <Variable name="D0"/>
                    <Variable name="D1"/>
                    <Variable name="C0"/>
                    <Variable name="C1"/>
                 </Variables>
                <TopVariable name="D0"/>
                <Branches>
                   <Branch>
                      <States><State name="no"/></States>
                      <Potential type="Tree/ADD">
                         <TopVariable name="C1"/>
                         <Branches>
                              <Branch>
                                  <Thresholds>
                                       <Threshold value="–Infinity"/>
                                       <Threshold value="0" belongsTo="Left"/>
                                  </Thresholds>
                                  <Potential type="MixtureOfExponentials">
                                      <Variables>
                                         <Variable name="C0"/>
                                         <Variable name="C1"/>
                                      </Variables>
                                      <Subpotentials>
                                         <Potential type="Exponential">
                                             <Potential type="Table">
                                                  <Values>3</Values>
                                              </Potential>
                                          </Potential>
                                          <Potential type="Exponential">
                                              <Potential type="Table">
                                                  <Values>–1</Values>
                                              </Potential>
                                              <NumericVariables>
                                                  <Variable name="C0"/>
                                                  <Variable name="C1"/>
                                              </NumericVariables>
                                              <Coefficients>4 –1</Coefficients>
                                           </Potential>
                                      </Subpotentials>
                                 </Potential>
                              </Branch>
                              <Branch>
                                    <Thresholds>
                                        <Threshold value="0" belongsTo="Left" />
                                        <Threshold value="+Infinity" />
                                     </Thresholds>
                                     <Potential type="MixtureOfExponentials">
                                        <Variables>
                                              <Variable name="C1"/>
                                              <Variable name="D1"/>
                                        </Variables>
                                        <Subpotentials>
                                              <Potential type="Exponential">
                                                 <Potential type="Table">
                                                      <Variables>
                                                          <Variable name="D1"/>
                                                      </Variables>
                                                      <Values>10  5</Values>
                                                 </Potential>
                                                 <NumericVariables>
                                                    <Variable name="C1"/>
                                                 </NumericVariables>
                                                 <Coefficients>0.25</Coefficients>
                                              </Potential>
                                         </Subpotentials>
                                     </Potential>
                              </Branch>
                         </Branches>
                      </Potential>
                   </Branch>
                  <Branch>
                        <States>
                            <State name="yes"/>
                         </States>
                         <Potential type="MixtureOfExponentials">
                                <Variables>
                                    <Variable name="C0"/>
                                </Variables>
                                <Subpotentials>
                                    <Potential type="Exponential">
                                        <Potential type="Table">
                                             <Values>0.3</Values>
                                        </Potential>
                                        <NumericVariables>
                                             <Variable name="C0"/>
                                        </NumericVariables>
                                        <Coefficients>1</Coefficients>
                                    </Potential>
                                   <Potential type="Exponential">
                                        <Potential type="Table">
                                             <Values>0.7</Values>
                                        </Potential>
                                   </Potential>
                                </Subpotentials>
                         </Potential>
                  </Branch>
                </Branches>
          </Potential>
      </Potentials>
    </ProbNet>
    <Policies />
    <InferenceOptions />
</ProbModelXML>
"""
        self.maxDiff = None
        self.reader_string = ProbModelXMLReader(string=string)
        self.reader_file = ProbModelXMLReader(path=six.StringIO(string))

    def test_comment(self):
        comment_expected = ("Student example model from Probabilistic Graphical Models: "
                            "Principles and Techniques by Daphne Koller")
        self.assertEqual(self.reader_string.probnet['Comment'], comment_expected)
        self.assertEqual(self.reader_file.probnet['Comment'], comment_expected)

    def test_variables(self):
        variables = {'difficulty':
                     {'Comment': None,
                      'Coordinates': {},
                      'role': 'Chance',
                      'type': 'FiniteState',
                      'States': {'difficult': {}, 'easy': {}}},
                     'intelligence':
                     {'Comment': None,
                      'Coordinates': {},
                      'role': 'Chance',
                      'type': 'FiniteState',
                      'States': {'smart': {}, 'dumb': {}}}}
        self.assertDictEqual(self.reader_string.probnet['Variables'], variables)
        self.assertDictEqual(self.reader_file.probnet['Variables'], variables)

    def test_edges(self):
        edge = {('grade', 'recommendation_letter'):
                {'directed': '1',
                 'Comment': 'Directed Edge from grade to recommendation_letter',
                 'Label': 'grad_to_reco'},
                ('intelligence', 'grade'):
                {'directed': '1',
                 'Comment': 'Directed Edge from intelligence to grade',
                 'Label': 'intel_to_grad'},
                ('difficulty', 'grade'):
                {'directed': '1',
                 'Comment': 'Directed Edge from difficulty to grade',
                 'Label': 'diff_to_grad'},
                ('intelligence', 'SAT'):
                {'directed': '1',
                 'Comment': 'Directed Edge from intelligence to SAT',
                 'Label': 'intel_to_sat'}}
        self.assertDictEqual(self.reader_string.probnet['edges'], edge)
        self.assertDictEqual(self.reader_file.probnet['edges'], edge)

    def test_additionalconstraints(self):
        additionalconstraints_expected = {'MaxNumParents':
                                          {'numParents': '5'}}
        self.assertDictEqual(self.reader_string.probnet['AdditionalConstraints'],
                             additionalconstraints_expected)
        self.assertDictEqual(self.reader_file.probnet['AdditionalConstraints'],
                             additionalconstraints_expected)

    def test_additionalproperties(self):
        additionalproperties_expected = {'elvira.title': 'X ray result'}
        self.assertDictEqual(self.reader_string.probnet['AdditionalProperties'],
                             additionalproperties_expected)
        self.assertDictEqual(self.reader_file.probnet['AdditionalProperties'],
                             additionalproperties_expected)

    def test_decisioncriteria(self):
        decisioncriteria_expected = {'effectiveness': {},
                                     'cost': {}}
        self.assertDictEqual(self.reader_string.probnet['DecisionCriteria'],
                             decisioncriteria_expected)
        self.assertDictEqual(self.reader_file.probnet['DecisionCriteria'],
                             decisioncriteria_expected)

    def test_potential(self):
        potential_expected = [{
            'role': 'Utility',
            'Variables': {'D0': ['D1', 'C0', 'C1']},
            'type': 'Tree/ADD',
            'UtilityVaribale': 'U1',
            'Branches': [{
                'Potential': {
                    'type': 'Tree/ADD',
                    'Branches': [{'Thresholds': [{'value': u'–Infinity'},
                                                 {'value': '0', 'belongsTo': 'Left'}],
                                  'Potential': {'Subpotentials': [{'Potential': {'type': 'Table',
                                                                                 'Values': '3'},
                                                                   'type': 'Exponential'},
                                                                  {'NumericVariables': ['C0', 'C1'],
                                                                   'Potential': {'type': 'Table',
                                                                                 'Values': u'–1'},
                                                                   'Coefficients': u'4 –1',
                                                                   'type': 'Exponential'}],
                                                'Variables': {'C0': ['C1']},
                                                'type': 'MixtureOfExponentials'}},
                                 {'Thresholds': [{'value': '0', 'belongsTo': 'Left'},
                                                 {'value': '+Infinity'}],
                                  'Potential': {'Subpotentials': [{'NumericVariables': ['C1'],
                                                                   'Potential': {'Variables': {'D1': []},
                                                                                 'type': 'Table',
                                                                                 'Values': '10  5'},
                                                                   'Coefficients': '0.25',
                                                                   'type': 'Exponential'}],
                                                'Variables': {'C1': ['D1']},
                                                'type': 'MixtureOfExponentials'}}],
                    'TopVariable': 'C1'},
                'States': [{'name': 'no'}]},
                {'Potential': {'Subpotentials': [{'NumericVariables': ['C0'],
                                                  'Potential': {'type': 'Table',
                                                                'Values': '0.3'},
                                                  'Coefficients': '1',
                                                  'type': 'Exponential'},
                                                 {'Potential': {'type': 'Table',
                                                                'Values': '0.7'},
                                                  'type': 'Exponential'}],
                               'Variables': {'C0': []},
                               'type': 'MixtureOfExponentials'},
                 'States': [{'name': 'yes'}]}],
            'TopVariable': 'D0'}]

        self.assertListEqual(self.reader_string.probnet['Potentials'],
                             potential_expected)
        self.assertListEqual(self.reader_file.probnet['Potentials'],
                             potential_expected)

    def test_get_model(self):
        string = """<ProbModelXML formatVersion="0.2.0">
  <ProbNet type="BayesianNetwork">
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
    <Variables>
      <Variable name="X-ray" type="finiteStates" role="chance" isInput="false">
        <Comment><![CDATA[Indica si el test de rayos X ha sido positivo]]></Comment>
        <Coordinates x="252" y="322" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="X" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Bronchitis" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="698" y="181" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="B" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Dyspnea" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="533" y="321" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="D" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="VisitToAsia" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="290" y="58" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="A" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Smoker" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="568" y="52" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="S" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="LungCancer" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="421" y="152" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="L" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="Tuberculosis" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="201" y="150" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="T" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
      <Variable name="TuberculosisOrCancer" type="finiteStates" role="chance" isInput="false">
        <Coordinates x="336" y="238" />
        <AdditionalProperties>
          <Property name="Relevance" value="7.0" />
          <Property name="Title" value="E" />
        </AdditionalProperties>
        <Precision>0.01</Precision>
        <States>
          <State name="no" />
          <State name="yes" />
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="true">
        <Variable name="Bronchitis" />
        <Variable name="Dyspnea" />
      </Link>
      <Link directed="true">
        <Variable name="VisitToAsia" />
        <Variable name="Tuberculosis" />
      </Link>
      <Link directed="true">
        <Variable name="Smoker" />
        <Variable name="Bronchitis" />
      </Link>
      <Link directed="true">
        <Variable name="Smoker" />
        <Variable name="LungCancer" />
      </Link>
      <Link directed="true">
        <Variable name="LungCancer" />
        <Variable name="TuberculosisOrCancer" />
      </Link>
      <Link directed="true">
        <Variable name="Tuberculosis" />
        <Variable name="TuberculosisOrCancer" />
      </Link>
      <Link directed="true">
        <Variable name="TuberculosisOrCancer" />
        <Variable name="Dyspnea" />
      </Link>
      <Link directed="true">
        <Variable name="TuberculosisOrCancer" />
        <Variable name="X-ray" />
      </Link>
    </Links>
    <Potentials>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="X-ray" />
          <Variable name="TuberculosisOrCancer" />
        </Variables>
        <Values>0.95 0.05 0.02 0.98</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Bronchitis" />
          <Variable name="Smoker" />
        </Variables>
        <Values>0.7 0.3 0.4 0.6</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Dyspnea" />
          <Variable name="TuberculosisOrCancer" />
          <Variable name="Bronchitis" />
        </Variables>
        <Values>0.9 0.1 0.3 0.7 0.2 0.8 0.1 0.9</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="VisitToAsia" />
        </Variables>
        <Values>0.99 0.01</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Smoker" />
        </Variables>
        <Values>0.5 0.5</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="LungCancer" />
          <Variable name="Smoker" />
        </Variables>
        <Values>0.99 0.01 0.9 0.1</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="Tuberculosis" />
          <Variable name="VisitToAsia" />
        </Variables>
        <Values>0.99 0.01 0.95 0.05</Values>
      </Potential>
      <Potential type="Table" role="conditionalProbability">
        <Variables>
          <Variable name="TuberculosisOrCancer" />
          <Variable name="LungCancer" />
          <Variable name="Tuberculosis" />
        </Variables>
        <Values>1.0 0.0 0.0 1.0 0.0 1.0 0.0 1.0</Values>
      </Potential>
    </Potentials>
    <AdditionalProperties>
      <Property name="VisualPrecision" value="0.0" />
      <Property name="KindOfGraph" value="directed" />
      <Property name="WhenChanged" value="19/08/99" />
      <Property name="Version" value="1.0" />
      <Property name="WhoChanged" value="Jose A. Gamez" />
    </AdditionalProperties>
  </ProbNet>
</ProbModelXML>"""
        self.maxDiff = None
        self.reader = ProbModelXMLReader(string=string)
        model = self.reader.get_model()
        edges_expected = [('VisitToAsia', 'Tuberculosis'),
                          ('LungCancer', 'TuberculosisOrCancer'),
                          ('Smoker', 'LungCancer'),
                          ('Smoker', 'Bronchitis'),
                          ('Tuberculosis', 'TuberculosisOrCancer'),
                          ('Bronchitis', 'Dyspnea'),
                          ('TuberculosisOrCancer', 'Dyspnea'),
                          ('TuberculosisOrCancer', 'X-ray')]
        node_expected = {'Smoker': {'States': {'no': {}, 'yes': {}},
                                    'role': 'chance',
                                    'type': 'finiteStates',
                                    'Coordinates': {'y': '52', 'x': '568'},
                                    'AdditionalProperties': {'Title': 'S', 'Relevance': '7.0'}},
                         'Bronchitis': {'States': {'no': {}, 'yes': {}},
                                        'role': 'chance',
                                        'type': 'finiteStates',
                                        'Coordinates': {'y': '181', 'x': '698'},
                                        'AdditionalProperties': {'Title': 'B', 'Relevance': '7.0'}},
                         'VisitToAsia': {'States': {'no': {}, 'yes': {}},
                                         'role': 'chance',
                                         'type': 'finiteStates',
                                         'Coordinates': {'y': '58', 'x': '290'},
                                         'AdditionalProperties': {'Title': 'A', 'Relevance': '7.0'}},
                         'Tuberculosis': {'States': {'no': {}, 'yes': {}},
                                          'role': 'chance',
                                          'type': 'finiteStates',
                                          'Coordinates': {'y': '150', 'x': '201'},
                                          'AdditionalProperties': {'Title': 'T', 'Relevance': '7.0'}},
                         'X-ray': {'States': {'no': {}, 'yes': {}},
                                   'role': 'chance',
                                   'AdditionalProperties': {'Title': 'X', 'Relevance': '7.0'},
                                   'Coordinates': {'y': '322', 'x': '252'},
                                   'Comment': 'Indica si el test de rayos X ha sido positivo',
                                   'type': 'finiteStates'},
                         'Dyspnea': {'States': {'no': {}, 'yes': {}},
                                     'role': 'chance',
                                     'type': 'finiteStates',
                                     'Coordinates': {'y': '321', 'x': '533'},
                                     'AdditionalProperties': {'Title': 'D', 'Relevance': '7.0'}},
                         'TuberculosisOrCancer': {'States': {'no': {}, 'yes': {}},
                                                  'role': 'chance',
                                                  'type': 'finiteStates',
                                                  'Coordinates': {'y': '238', 'x': '336'},
                                                  'AdditionalProperties': {'Title': 'E', 'Relevance': '7.0'}},
                         'LungCancer': {'States': {'no': {}, 'yes': {}},
                                        'role': 'chance',
                                        'type': 'finiteStates',
                                        'Coordinates': {'y': '152', 'x': '421'},
                                        'AdditionalProperties': {'Title': 'L', 'Relevance': '7.0'}}}
        edge_expected = {'LungCancer': {'TuberculosisOrCancer': {'weight': None,
                                                                 'directed': 'true'}},
                         'Smoker': {'LungCancer': {'weight': None,
                                                   'directed': 'true'},
                                    'Bronchitis': {'weight': None,
                                                   'directed': 'true'}},
                         'Dyspnea': {},
                         'X-ray': {},
                         'VisitToAsia': {'Tuberculosis': {'weight': None,
                                                          'directed': 'true'}},
                         'TuberculosisOrCancer': {'X-ray': {'weight': None,
                                                            'directed': 'true'},
                                                  'Dyspnea': {'weight': None,
                                                              'directed': 'true'}},
                         'Bronchitis': {'Dyspnea': {'weight': None,
                                                    'directed': 'true'}},
                         'Tuberculosis': {'TuberculosisOrCancer': {'weight': None,
                                                                   'directed': 'true'}}}

        cpds_expected = [np.array([[0.95, 0.05], [0.02, 0.98]]),
                         np.array([[0.7, 0.3], [0.4,  0.6]]),
                         np.array([[0.9, 0.1,  0.3,  0.7], [0.2,  0.8,  0.1,  0.9]]),
                         np.array([[0.99], [0.01]]),
                         np.array([[0.5], [0.5]]),
                         np.array([[0.99, 0.01], [0.9, 0.1]]),
                         np.array([[0.99, 0.01], [0.95, 0.05]]),
                         np.array([[1, 0, 0, 1], [0, 1, 0, 1]])]
        for cpd_index in range(0, len(cpds_expected)):
            np_test.assert_array_equal(model.get_cpds()[cpd_index].get_values(),
                                       cpds_expected[cpd_index])
        self.assertDictEqual(model.node, node_expected)
        self.assertDictEqual(model.edge, edge_expected)
        self.assertListEqual(sorted(model.edges()), sorted(edges_expected))


class TestProbModelXMLWriter(unittest.TestCase):
    def setUp(self):
        self.model_data = {'probnet':
                           {'type': 'BayesianNetwork',
                            'Language': 'English',
                            'AdditionalConstraints': {'MaxNumParents':
                                                      {'numParents': '5'}},
                            'AdditionalProperties': {'elvira.title': 'X ray result'},
                            'DecisionCriteria': {'effectiveness': {},
                                                 'cost': {}},
                            'Variables': {'difficulty':
                                          {'type': 'FiniteState',
                                           'role': 'Chance',
                                           'States': {'difficult': {}, 'easy': {}},
                                           'Comment': None,
                                           'Coordinates': {}},
                                          'intelligence':
                                          {'type': 'FiniteState',
                                           'role': 'Chance',
                                           'States': {'smart': {}, 'dumb': {}},
                                           'Comment': None,
                                           'Coordinates': {}}},
                            'Comment': 'Student example model from Probabilistic Graphical Models: '
                                       'Principles and Techniques by Daphne Koller',
                            'edges': {('difficulty', 'grade'):
                                      {'directed': '1',
                                       'Label': 'diff_to_grad',
                                       'Comment': 'Directed Edge from difficulty to grade'},
                                      ('intelligence', 'grade'):
                                      {'directed': '1',
                                       'Label': 'intel_to_grad',
                                       'Comment': 'Directed Edge from intelligence to grade'},
                                      ('intelligence', 'SAT'):
                                      {'directed': '1',
                                       'Label': 'intel_to_sat',
                                       'Comment': 'Directed Edge from intelligence to SAT'},
                                      ('grade', 'recommendation_letter'):
                                      {'directed': '1',
                                       'Label': 'grad_to_reco',
                                       'Comment': 'Directed Edge from grade to recommendation_letter'}},
                            'Potentials': [{'role': 'Utility',
                                            'Variables': {'D0': ['D1', 'C0', 'C1']},
                                            'type': 'Tree/ADD',
                                            'UtilityVaribale': 'U1',
                                            'Branches': [{
                                                'Potential': {
                                                    'type': 'Tree/ADD',
                                                    'Branches': [{
                                                        'Thresholds': [{'value': '-Infinity'},
                                                                       {'value': '0', 'belongsTo': 'Left'}],
                                                        'Potential': {'Subpotentials': [
                                                            {'Potential': {'type': 'Table',
                                                                           'Values': '3'},
                                                             'type': 'Exponential'},
                                                            {'NumericVariables': ['C0', 'C1'],
                                                             'Potential': {'type': 'Table',
                                                                           'Values': '-1'},
                                                             'Coefficients': '4 -1',
                                                             'type': 'Exponential'}],
                                                            'Variables': {'C0': ['C1']},
                                                            'type': 'MixtureOfExponentials'}},
                                                        {'Thresholds': [{'value': '0', 'belongsTo': 'Left'},
                                                                        {'value': '+Infinity'}],
                                                         'Potential': {'Subpotentials': [
                                                             {'NumericVariables': ['C1'],
                                                              'Potential': {'Variables': {'D1': []},
                                                                            'type': 'Table',
                                                                            'Values': '10  5'},
                                                              'Coefficients': '0.25',
                                                              'type': 'Exponential'}],
                                                             'Variables': {'C1': ['D1']},
                                                             'type': 'MixtureOfExponentials'}}],
                                                    'TopVariable': 'C1'},
                                                'States': [{'name': 'no'}]},
                                                {'Potential': {'Subpotentials': [
                                                    {'NumericVariables': ['C0'],
                                                     'Potential': {'type': 'Table',
                                                                   'Values': '0.3'},
                                                     'Coefficients': '1',
                                                     'type': 'Exponential'},
                                                    {'Potential': {'type': 'Table',
                                                                   'Values': '0.7'},
                                                     'type': 'Exponential'}],
                                                    'Variables': {'C0': []},
                                                    'type': 'MixtureOfExponentials'},
                                                    'States': [{'name': 'yes'}]}],
                                            'TopVariable': 'D0'}]}}

        self.maxDiff = None
        self.writer = ProbModelXMLWriter(model_data=self.model_data)

    def test_file(self):
        self.expected_xml = etree.XML("""<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="difficulty" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="difficult">
            <AdditionalProperties/>
          </State>
          <State name="easy">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="intelligence" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="dumb">
            <AdditionalProperties/>
          </State>
          <State name="smart">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="1" var1="difficulty" var2="grade">
        <Comment>Directed Edge from difficulty to grade</Comment>
        <Label>diff_to_grad</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="grade" var2="recommendation_letter">
        <Comment>Directed Edge from grade to recommendation_letter</Comment>
        <Label>grad_to_reco</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="SAT">
        <Comment>Directed Edge from intelligence to SAT</Comment>
        <Label>intel_to_sat</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="grade">
        <Comment>Directed Edge from intelligence to grade</Comment>
        <Label>intel_to_grad</Label>
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="Utility" type="Tree/ADD">
        <Variables>
          <Variable name="D0"/>
          <Variable name="C0"/>
          <Variable name="C1"/>
          <Variable name="D1"/>
        </Variables>
        <TopVariable name="D0"/>
        <Branches>
          <Branch>
            <States>
              <State name="no"/>
            </States>
            <Potential type="Tree/ADD">
              <TopVariable name="C1"/>
              <Branches>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C0"/>
                      <Variable name="C1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Potential type="Table">
                          <Values>3</Values>
                        </Potential>
                      </Potential>
                      <Potential type="Exponential">
                        <Coefficients>4 -1</Coefficients>
                        <Potential type="Table">
                          <Values>-1</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C0"/>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold value="-Infinity"/>
                    <Threshold belongsTo="Left" value="0"/>
                  </Thresholds>
                </Branch>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C1"/>
                      <Variable name="D1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Coefficients>0.25</Coefficients>
                        <Potential type="Table">
                          <Variables>
                            <Variable name="D1"/>
                          </Variables>
                          <Values>10  5</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold belongsTo="Left" value="0"/>
                    <Threshold value="+Infinity"/>
                  </Thresholds>
                </Branch>
              </Branches>
            </Potential>
          </Branch>
          <Branch>
            <States>
              <State name="yes"/>
            </States>
            <Potential type="MixtureOfExponentials">
              <Variables>
                <Variable name="C0"/>
              </Variables>
              <Subpotentials>
                <Potential type="Exponential">
                  <Coefficients>1</Coefficients>
                  <Potential type="Table">
                    <Values>0.3</Values>
                  </Potential>
                  <NumericVariables>
                    <Variable name="C0"/>
                  </NumericVariables>
                </Potential>
                <Potential type="Exponential">
                  <Potential type="Table">
                    <Values>0.7</Values>
                  </Potential>
                </Potential>
              </Subpotentials>
            </Potential>
          </Branch>
        </Branches>
      </Potential>
    </Potentials>
    <AdditionalConstraints>
      <Constraint name="MaxNumParents">
        <Argument name="numParents" value="5"/>
      </Constraint>
    </AdditionalConstraints>
    <Language>English</Language>
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
  </ProbNet>
  <AdditionalProperties>
    <Property name="elvira.title" value="X ray result"/>
  </AdditionalProperties>
  <DecisionCriteria>
    <Criterion name="cost">
      <AdditionalProperties/>
    </Criterion>
    <Criterion name="effectiveness">
      <AdditionalProperties/>
    </Criterion>
  </DecisionCriteria>
</ProbModelXML>""")
        self.assertEqual(str(self.writer.__str__()[:-1]), str(etree.tostring(self.expected_xml)))

    def test_write_file(self):
        self.expected_xml = etree.XML("""<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="difficulty" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="difficult">
            <AdditionalProperties/>
          </State>
          <State name="easy">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="intelligence" role="Chance" type="FiniteState">
        <Comment/>
        <Coordinates/>
        <AdditionalProperties/>
        <States>
          <State name="dumb">
            <AdditionalProperties/>
          </State>
          <State name="smart">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="1" var1="difficulty" var2="grade">
        <Comment>Directed Edge from difficulty to grade</Comment>
        <Label>diff_to_grad</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="grade" var2="recommendation_letter">
        <Comment>Directed Edge from grade to recommendation_letter</Comment>
        <Label>grad_to_reco</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="SAT">
        <Comment>Directed Edge from intelligence to SAT</Comment>
        <Label>intel_to_sat</Label>
        <AdditionalProperties/>
      </Link>
      <Link directed="1" var1="intelligence" var2="grade">
        <Comment>Directed Edge from intelligence to grade</Comment>
        <Label>intel_to_grad</Label>
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="Utility" type="Tree/ADD">
        <Variables>
          <Variable name="D0"/>
          <Variable name="C0"/>
          <Variable name="C1"/>
          <Variable name="D1"/>
        </Variables>
        <TopVariable name="D0"/>
        <Branches>
          <Branch>
            <States>
              <State name="no"/>
            </States>
            <Potential type="Tree/ADD">
              <TopVariable name="C1"/>
              <Branches>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C0"/>
                      <Variable name="C1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Potential type="Table">
                          <Values>3</Values>
                        </Potential>
                      </Potential>
                      <Potential type="Exponential">
                        <Coefficients>4 -1</Coefficients>
                        <Potential type="Table">
                          <Values>-1</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C0"/>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold value="-Infinity"/>
                    <Threshold belongsTo="Left" value="0"/>
                  </Thresholds>
                </Branch>
                <Branch>
                  <Potential type="MixtureOfExponentials">
                    <Variables>
                      <Variable name="C1"/>
                      <Variable name="D1"/>
                    </Variables>
                    <Subpotentials>
                      <Potential type="Exponential">
                        <Coefficients>0.25</Coefficients>
                        <Potential type="Table">
                          <Variables>
                            <Variable name="D1"/>
                          </Variables>
                          <Values>10  5</Values>
                        </Potential>
                        <NumericVariables>
                          <Variable name="C1"/>
                        </NumericVariables>
                      </Potential>
                    </Subpotentials>
                  </Potential>
                  <Thresholds>
                    <Threshold belongsTo="Left" value="0"/>
                    <Threshold value="+Infinity"/>
                  </Thresholds>
                </Branch>
              </Branches>
            </Potential>
          </Branch>
          <Branch>
            <States>
              <State name="yes"/>
            </States>
            <Potential type="MixtureOfExponentials">
              <Variables>
                <Variable name="C0"/>
              </Variables>
              <Subpotentials>
                <Potential type="Exponential">
                  <Coefficients>1</Coefficients>
                  <Potential type="Table">
                    <Values>0.3</Values>
                  </Potential>
                  <NumericVariables>
                    <Variable name="C0"/>
                  </NumericVariables>
                </Potential>
                <Potential type="Exponential">
                  <Potential type="Table">
                    <Values>0.7</Values>
                  </Potential>
                </Potential>
              </Subpotentials>
            </Potential>
          </Branch>
        </Branches>
      </Potential>
    </Potentials>
    <AdditionalConstraints>
      <Constraint name="MaxNumParents">
        <Argument name="numParents" value="5"/>
      </Constraint>
    </AdditionalConstraints>
    <Language>English</Language>
    <Comment>Student example model from Probabilistic Graphical Models: Principles and Techniques by Daphne Koller</Comment>
  </ProbNet>
  <AdditionalProperties>
    <Property name="elvira.title" value="X ray result"/>
  </AdditionalProperties>
  <DecisionCriteria>
    <Criterion name="cost">
      <AdditionalProperties/>
    </Criterion>
    <Criterion name="effectiveness">
      <AdditionalProperties/>
    </Criterion>
  </DecisionCriteria>
</ProbModelXML>""")
        self.writer.write_file("test_xml.pgmx")
        with open("test_xml.pgmx", "r") as myfile:
            data = myfile.read()
        self.assertEqual(str(self.writer.__str__()[:-1]), str(etree.tostring(self.expected_xml)))
        self.assertEqual(str(data), str(etree.tostring(self.expected_xml).decode('utf-8')))

    def tearDown(self):
        try:
            os.remove('test_xml.pgmx')
        except OSError:
            pass

class TestProbModelXMLmethods(unittest.TestCase):
    def setUp(self):
        edges_list = [('VisitToAsia', 'Tuberculosis'),
                      ('LungCancer', 'TuberculosisOrCancer'),
                      ('Smoker', 'LungCancer'),
                      ('Smoker', 'Bronchitis'),
                      ('Tuberculosis', 'TuberculosisOrCancer'),
                      ('Bronchitis', 'Dyspnea'),
                      ('TuberculosisOrCancer', 'Dyspnea'),
                      ('TuberculosisOrCancer', 'X-ray')]
        nodes = {'Smoker': {'States': {'no': {}, 'yes': {}},
                            'role': 'chance',
                            'type': 'finiteStates',
                            'Coordinates': {'y': '52', 'x': '568'},
                            'AdditionalProperties': {'Title': 'S', 'Relevance': '7.0'}},
                 'Bronchitis': {'States': {'no': {}, 'yes': {}},
                                'role': 'chance',
                                'type': 'finiteStates',
                                'Coordinates': {'y': '181', 'x': '698'},
                                'AdditionalProperties': {'Title': 'B', 'Relevance': '7.0'}},
                 'VisitToAsia': {'States': {'no': {}, 'yes': {}},
                                 'role': 'chance',
                                 'type': 'finiteStates',
                                 'Coordinates': {'y': '58', 'x': '290'},
                                 'AdditionalProperties': {'Title': 'A', 'Relevance': '7.0'}},
                 'Tuberculosis': {'States': {'no': {}, 'yes': {}},
                                  'role': 'chance',
                                  'type': 'finiteStates',
                                  'Coordinates': {'y': '150', 'x': '201'},
                                  'AdditionalProperties': {'Title': 'T', 'Relevance': '7.0'}},
                 'X-ray': {'States': {'no': {}, 'yes': {}},
                           'role': 'chance',
                           'AdditionalProperties': {'Title': 'X', 'Relevance': '7.0'},
                           'Coordinates': {'y': '322', 'x': '252'},
                           'Comment': 'Indica si el test de rayos X ha sido positivo',
                           'type': 'finiteStates'},
                 'Dyspnea': {'States': {'no': {}, 'yes': {}},
                             'role': 'chance',
                             'type': 'finiteStates',
                             'Coordinates': {'y': '321', 'x': '533'},
                             'AdditionalProperties': {'Title': 'D', 'Relevance': '7.0'}},
                 'TuberculosisOrCancer': {'States': {'no': {}, 'yes': {}},
                                          'role': 'chance',
                                          'type': 'finiteStates',
                                          'Coordinates': {'y': '238', 'x': '336'},
                                          'AdditionalProperties': {'Title': 'E', 'Relevance': '7.0'}},
                 'LungCancer': {'States': {'no': {}, 'yes': {}},
                                'role': 'chance',
                                'type': 'finiteStates',
                                'Coordinates': {'y': '152', 'x': '421'},
                                'AdditionalProperties': {'Title': 'L', 'Relevance': '7.0'}}}
        edges = {'LungCancer': {'TuberculosisOrCancer': {'directed': 'true'}},
                 'Smoker': {'LungCancer': {'directed': 'true'},
                            'Bronchitis': {'directed': 'true'}},
                 'Dyspnea': {},
                 'X-ray': {},
                 'VisitToAsia': {'Tuberculosis': {'directed': 'true'}},
                 'TuberculosisOrCancer': {'X-ray': {'directed': 'true'},
                                          'Dyspnea': {'directed': 'true'}},
                 'Bronchitis': {'Dyspnea': {'directed': 'true'}},
                 'Tuberculosis': {'TuberculosisOrCancer': {'directed': 'true'}}}

        cpds = [{'Values': np.array([[0.95, 0.05], [0.02, 0.98]]),
                 'Variables': {'X-ray': ['TuberculosisOrCancer']}},
                {'Values': np.array([[0.7, 0.3], [0.4,  0.6]]),
                 'Variables': {'Bronchitis': ['Smoker']}},
                {'Values':  np.array([[0.9, 0.1,  0.3,  0.7], [0.2,  0.8,  0.1,  0.9]]),
                 'Variables': {'Dyspnea': ['TuberculosisOrCancer', 'Bronchitis']}},
                {'Values': np.array([[0.99], [0.01]]),
                 'Variables': {'VisitToAsia': []}},
                {'Values': np.array([[0.5], [0.5]]),
                 'Variables': {'Smoker': []}},
                {'Values': np.array([[0.99, 0.01], [0.9, 0.1]]),
                 'Variables': {'LungCancer': ['Smoker']}},
                {'Values': np.array([[0.99, 0.01], [0.95, 0.05]]),
                 'Variables': {'Tuberculosis': ['VisitToAsia']}},
                {'Values': np.array([[1, 0, 0, 1], [0, 1, 0, 1]]),
                 'Variables': {'TuberculosisOrCancer': ['LungCancer', 'Tuberculosis']}}]
        self.model = BayesianModel(edges_list)
        for node in nodes:
            self.model.node[node] = nodes[node]
        for edge in edges:
            self.model.edge[edge] = edges[edge]

        tabular_cpds = []
        for cpd in cpds:
            var = list(cpd['Variables'].keys())[0]
            evidence = cpd['Variables'][var]
            values = cpd['Values']
            states = len(nodes[var]['States'])
            evidence_card = [len(nodes[evidence_var]['States'])
                             for evidence_var in evidence]
            tabular_cpds.append(
                TabularCPD(var, states, values, evidence, evidence_card))
        self.maxDiff = None
        self.model.add_cpds(*tabular_cpds)

    def test_get_probmodel_data(self):
        model_data = get_probmodel_data(self.model)
        xmlfile = ProbModelXMLWriter(model_data)
        with open('pgmpy/tests/test_readwrite/testdata/test_probmodelxml_data.json') as data_file:
            model_data_expected = json.load(data_file)
        xmlfile_expected = etree.XML("""<ProbModelXML formatVersion="1.0">
  <ProbNet type="BayesianNetwork">
    <Variables>
      <Variable name="Bronchitis" role="chance" type="finiteStates">
        <Coordinates x="698" y="181"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="B"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Dyspnea" role="chance" type="finiteStates">
        <Coordinates x="533" y="321"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="D"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="LungCancer" role="chance" type="finiteStates">
        <Coordinates x="421" y="152"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="L"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Smoker" role="chance" type="finiteStates">
        <Coordinates x="568" y="52"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="S"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="Tuberculosis" role="chance" type="finiteStates">
        <Coordinates x="201" y="150"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="T"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="TuberculosisOrCancer" role="chance" type="finiteStates">
        <Coordinates x="336" y="238"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="E"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="VisitToAsia" role="chance" type="finiteStates">
        <Coordinates x="290" y="58"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="A"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
      <Variable name="X-ray" role="chance" type="finiteStates">
        <Comment>Indica si el test de rayos X ha sido positivo</Comment>
        <Coordinates x="252" y="322"/>
        <Property name="Relevance" value="7.0"/>
        <Property name="Title" value="X"/>
        <States>
          <State name="no">
            <AdditionalProperties/>
          </State>
          <State name="yes">
            <AdditionalProperties/>
          </State>
        </States>
      </Variable>
    </Variables>
    <Links>
      <Link directed="true" var1="Bronchitis" var2="Dyspnea">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="LungCancer" var2="TuberculosisOrCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Smoker" var2="Bronchitis">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Smoker" var2="LungCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="Tuberculosis" var2="TuberculosisOrCancer">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="TuberculosisOrCancer" var2="Dyspnea">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="TuberculosisOrCancer" var2="X-ray">
        <AdditionalProperties/>
      </Link>
      <Link directed="true" var1="VisitToAsia" var2="Tuberculosis">
        <AdditionalProperties/>
      </Link>
    </Links>
    <Potentials>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="X-ray"/>
          <Variable name="TuberculosisOrCancer"/>
        </Variables>
        <Values>0.95 0.05 0.02 0.98 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Bronchitis"/>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.7 0.3 0.4 0.6 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Dyspnea"/>
          <Variable name="Bronchitis"/>
          <Variable name="TuberculosisOrCancer"/>
        </Variables>
        <Values>0.9 0.1 0.3 0.7 0.2 0.8 0.1 0.9 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="VisitToAsia"/>
        </Variables>
        <Values>0.99 0.01 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.5 0.5 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="LungCancer"/>
          <Variable name="Smoker"/>
        </Variables>
        <Values>0.99 0.01 0.9 0.1 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="Tuberculosis"/>
          <Variable name="VisitToAsia"/>
        </Variables>
        <Values>0.99 0.01 0.95 0.05 </Values>
      </Potential>
      <Potential role="conditionalProbability" type="Table">
        <Variables>
          <Variable name="TuberculosisOrCancer"/>
          <Variable name="LungCancer"/>
          <Variable name="Tuberculosis"/>
        </Variables>
        <Values>1.0 0.0 0.0 1.0 0.0 1.0 0.0 1.0 </Values>
      </Potential>
    </Potentials>
    <AdditionalConstraints/>
    <AdditionalProperties/>
    <DecisionCriteria/>
  </ProbNet>
</ProbModelXML>""")
        self.assertDictEqual(model_data, model_data_expected)
        self.assertEqual(str(xmlfile.__str__()[:-1]), str(etree.tostring(xmlfile_expected)))



#!/usr/bin/env python3

import unittest

from pgmpy.base import DirectedGraph
import pgmpy.tests.help_functions as hf


class TestDirectedGraphCreation(unittest.TestCase):
    def setUp(self):
        self.graph = DirectedGraph()

    def test_class_init_without_data(self):
        self.assertIsInstance(self.graph, DirectedGraph)

    def test_class_init_with_data_string(self):
        self.graph = DirectedGraph([('a', 'b'), ('b', 'c')])
        self.assertListEqual(sorted(self.graph.nodes()), ['a', 'b', 'c'])
        self.assertListEqual(hf.recursive_sorted(self.graph.edges()),
                             [['a', 'b'], ['b', 'c']])

    def test_add_node_string(self):
        self.graph.add_node('a')
        self.assertListEqual(self.graph.nodes(), ['a'])

    def test_add_node_nonstring(self):
        self.graph.add_node(1)

    def test_add_nodes_from_string(self):
        self.graph.add_nodes_from(['a', 'b', 'c', 'd'])
        self.assertListEqual(sorted(self.graph.nodes()), ['a', 'b', 'c', 'd'])

    def test_add_nodes_from_non_string(self):
        self.graph.add_nodes_from([1, 2, 3, 4])

    def test_add_node_weight(self):
        self.graph.add_node('weighted_a', 0.3)
        self.assertEqual(self.graph.node['weighted_a']['weight'], 0.3)

    def test_add_nodes_from_weight(self):
        self.graph.add_nodes_from(['weighted_b', 'weighted_c'], [0.5, 0.6])
        self.assertEqual(self.graph.node['weighted_b']['weight'], 0.5)
        self.assertEqual(self.graph.node['weighted_c']['weight'], 0.6)

        self.graph.add_nodes_from(['e', 'f'])
        self.assertEqual(self.graph.node['e']['weight'], None)
        self.assertEqual(self.graph.node['f']['weight'], None)

    def test_add_edge_string(self):
        self.graph.add_edge('d', 'e')
        self.assertListEqual(sorted(self.graph.nodes()), ['d', 'e'])
        self.assertListEqual(self.graph.edges(), [('d', 'e')])
        self.graph.add_nodes_from(['a', 'b', 'c'])
        self.graph.add_edge('a', 'b')
        self.assertListEqual(hf.recursive_sorted(self.graph.edges()),
                             [['a', 'b'], ['d', 'e']])

    def test_add_edge_nonstring(self):
        self.graph.add_edge(1, 2)

    def test_add_edges_from_string(self):
        self.graph.add_edges_from([('a', 'b'), ('b', 'c')])
        self.assertListEqual(sorted(self.graph.nodes()), ['a', 'b', 'c'])
        self.assertListEqual(hf.recursive_sorted(self.graph.edges()),
                             [['a', 'b'], ['b', 'c']])
        self.graph.add_nodes_from(['d', 'e', 'f'])
        self.graph.add_edges_from([('d', 'e'), ('e', 'f')])
        self.assertListEqual(sorted(self.graph.nodes()),
                             ['a', 'b', 'c', 'd', 'e', 'f'])
        self.assertListEqual(hf.recursive_sorted(self.graph.edges()),
                             hf.recursive_sorted([('a', 'b'), ('b', 'c'),
                                                  ('d', 'e'), ('e', 'f')]))

    def test_add_edges_from_nonstring(self):
        self.graph.add_edges_from([(1, 2), (2, 3)])

    def test_add_edge_weight(self):
        self.graph.add_edge('a', 'b', weight=0.3)
        self.assertEqual(self.graph.edge['a']['b']['weight'], 0.3)

    def test_add_edges_from_weight(self):
        self.graph.add_edges_from([('b', 'c'), ('c', 'd')], weights=[0.5, 0.6])
        self.assertEqual(self.graph.edge['b']['c']['weight'], 0.5)
        self.assertEqual(self.graph.edge['c']['d']['weight'], 0.6)

        self.graph.add_edges_from([('e', 'f')])
        self.assertEqual(self.graph.edge['e']['f']['weight'], None)

    def test_update_node_parents_bm_constructor(self):
        self.graph = DirectedGraph([('a', 'b'), ('b', 'c')])
        self.assertListEqual(self.graph.predecessors('a'), [])
        self.assertListEqual(self.graph.predecessors('b'), ['a'])
        self.assertListEqual(self.graph.predecessors('c'), ['b'])

    def test_update_node_parents(self):
        self.graph.add_nodes_from(['a', 'b', 'c'])
        self.graph.add_edges_from([('a', 'b'), ('b', 'c')])
        self.assertListEqual(self.graph.predecessors('a'), [])
        self.assertListEqual(self.graph.predecessors('b'), ['a'])
        self.assertListEqual(self.graph.predecessors('c'), ['b'])

    def test_get_leaves(self):
        self.graph.add_edges_from([('A', 'B'), ('B', 'C'), ('B', 'D'),
                                   ('D', 'E'), ('D', 'F'), ('A', 'G')])
        self.assertEqual(sorted(self.graph.get_leaves()),
                         sorted(['C', 'G', 'E', 'F']))

    def test_get_roots(self):
        self.graph.add_edges_from([('A', 'B'), ('B', 'C'), ('B', 'D'),
                                   ('D', 'E'), ('D', 'F'), ('A', 'G')])
        self.assertEqual(['A'], self.graph.get_roots())
        self.graph.add_edge('H', 'G')
        self.assertEqual(sorted(['A', 'H']), sorted(self.graph.get_roots()))

    def tearDown(self):
        del self.graph


class TestDirectedGraphMoralization(unittest.TestCase):
    def setUp(self):
        self.graph = DirectedGraph()
        self.graph.add_edges_from([('diff', 'grade'), ('intel', 'grade')])

    def test_get_parents(self):
        self.assertListEqual(sorted(self.graph.get_parents('grade')),
                             ['diff', 'intel'])

    def test_moralize(self):
        moral_graph = self.graph.moralize()
        self.assertListEqual(hf.recursive_sorted(moral_graph.edges()),
                             [['diff', 'grade'], ['diff', 'intel'],
                              ['grade', 'intel']])

    def tearDown(self):
        del self.graph



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np

from pgmpy.factors.distributions import BaseDistribution
from pgmpy.factors.distributions import GaussianDistribution


class CanonicalDistribution(BaseDistribution):
    u"""
    The intermediate factors in a Gaussian network can be described
    compactly using a simple parametric representation called the
    canonical form. This representation is closed under the basic
    operations used in inference: factor product, factor division,
    factor reduction, and marginalization. Thus, we define this
    CanonicalDistribution class that allows the inference process to be
    performed on joint Gaussian networks.

    A canonical form C (X; K,h,g) is defined as

    C (X; K,h,g) = exp( ((-1/2) * X.T * K * X) + (h.T * X) + g)

    Reference
    ---------
    Probabilistic Graphical Models, Principles and Techniques,
    Daphne Koller and Nir Friedman, Section 14.2, Chapter 14.

    """
    def __init__(self, variables, K, h, g):
        """
        Parameters
        ----------
        variables: list or array-like
        The variables for wich the distribution is defined.

        K: n x n, 2-d array-like

        h : n x 1, array-like

        g : int, float

        pdf: function
        The probability density function of the distribution.

        The terms K, h and g are defined parameters for canonical
        factors representation.

        Examples
        --------
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi = CanonicalDistribution(['X', 'Y'], np.array([[1, -1], [-1, 1]]),
                                  np.array([[1], [-1]]), -3)
        >>> phi.variables
        ['X', 'Y']

        >>> phi.K
        array([[1, -1],
               [-1, 1]])

        >>> phi.h
        array([[1],
               [-1]])

        >>> phi.g
        -3

        """
        no_of_var = len(variables)

        if len(h) != no_of_var:
            raise ValueError("Length of h parameter vector must be equal to "
                             "the number of variables.")

        self.variables = variables
        self.h = np.asarray(np.reshape(h, (no_of_var, 1)), dtype=float)
        self.g = g
        self.K = np.asarray(K, dtype=float)

        if self.K.shape != (no_of_var, no_of_var):
            raise ValueError("The K matrix should be a square matrix with "
                             "order equal to the number of variables. Got: "
                             "{got_shape}, Expected: {exp_shape}".format(
                                 got_shape=self.K.shape,
                                 exp_shape=(no_of_var, no_of_var)))

    @property
    def pdf(self):
        def fun(*args):
            x = np.array(args)
            return np.exp(self.g + np.dot(x, self.h)[0] - 0.5 * np.dot(x.T, np.dot(self.K, x)))
        return fun

    def assignment(self, *x):
        """
        Returns the probability value of the PDF at the given parameter values.
        Parameters
        ----------
        *x: values of all variables of this distribution,
            collective defining a point at which the probability value is to be computed.
        Returns
        -------
        float: The probability value at the point.
        Examples
        --------
        >>> from pgmpy.factors.distributions import GaussianDistribution
        >>> dist = GaussianDistribution(variables=['x1', 'x2'],
        ...                             mean=[[0], [0]],
        ...                             cov=[[1, 0], [0, 1]])
        >>> dist.assignment(0, 0)
        0.15915494309189535
        """
        return self.pdf(*x)

    def copy(self):
        """
        Makes a copy of the factor.

        Returns
        -------
        CanonicalDistribution object: Copy of the factor

        Examples
        --------
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi = CanonicalDistribution(['X', 'Y'], np.array([[1, -1], [-1, 1]]),
                                  np.array([[1], [-1]]), -3)
        >>> phi.variables
        ['X', 'Y']

        >>> phi.K
        array([[1, -1],
               [-1, 1]])

        >>> phi.h
        array([[1],
               [-1]])

        >>> phi.g
        -3

        >>> phi2 = phi.copy()

        >>> phi2.variables
        ['X', 'Y']

        >>> phi2.K
        array([[1, -1],
               [-1, 1]])

        >>> phi2.h
        array([[1],
               [-1]])

        >>> phi2.g
        -3

        """
        copy_factor = CanonicalDistribution(self.variables, self.K.copy(),
                                            self.h.copy(), self.g)

        return copy_factor

    def to_joint_gaussian(self):
        """
        Return an equivalent Joint Gaussian Distribution.

        Examples
        --------

        >>> import numpy as np
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi = CanonicalDistribution(['x1', 'x2'], np.array([[3, -2], [-2, 4]]),
                                  np.array([[5], [-1]]), 1)
        >>> jgd = phi.to_joint_gaussian()
        >>> jgd.variables
        ['x1', 'x2']
        >>> jgd.covariance
        array([[ 0.5  ,  0.25 ],
               [ 0.25 ,  0.375]])
        >>> jgd.mean
        array([[ 2.25 ],
               [ 0.875]])

        """
        covariance = np.linalg.inv(self.K)
        mean = np.dot(covariance, self.h)

        return GaussianDistribution(self.variables, mean, covariance)

    def reduce(self, values, inplace=True):
        """
        Reduces the distribution to the context of the given variable values.

        Let C(X,Y ; K, h, g) be some canonical form over X,Y where,

        k = [[K_XX, K_XY],        ;       h = [[h_X],
             [K_YX, K_YY]]                     [h_Y]]

        The formula for the obtained conditional distribution for setting
        Y = y is given by,

        .. math:: K' = K_{XX}
        .. math:: h' = h_X - K_{XY} * y
        .. math:: g' = g + {h^T}_Y * y - 0.5 * y^T * K_{YY} * y


        Parameters
        ----------
        values: list, array-like
            A list of tuples of the form (variable name, variable value).

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new CaninicalFactor object.

        Returns
        -------
        CanonicalDistribution or None:
                if inplace=True (default) returns None
                if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi = CanonicalDistribution(['X1', 'X2', 'X3'],
        ...                       np.array([[1, -1, 0], [-1, 4, -2], [0, -2, 4]]),
        ...                       np.array([[1], [4], [-1]]), -2)
        >>> phi.variables
        ['X1', 'X2', 'X3']

        >>> phi.K
        array([[ 1., -1.],
               [-1.,  3.]])

        >>> phi.h
        array([[ 1. ],
               [ 3.5]])

        >>> phi.g
        -2

        >>> phi.reduce([('X3', 0.25)])

        >>> phi.variables
        ['X1', 'X2']

        >>> phi.K
        array([[ 1, -1],
               [-1,  4]])

        >>> phi.h
        array([[ 1. ],
               [ 4.5]])

        >>> phi.g
        -2.375
        """
        if not isinstance(values, (list, tuple, np.ndarray)):
            raise TypeError("variables: Expected type list or array-like, "
                            "got type {var_type}".format(var_type=type(values)))

        if not all([var in self.variables for var, value in values]):
            raise ValueError("Variable not in scope.")

        phi = self if inplace else self.copy()

        var_to_reduce = [var for var, value in values]

        # index_to_keep -> j vector
        index_to_keep = [self.variables.index(var) for var in self.variables
                         if var not in var_to_reduce]
        # index_to_reduce -> i vector
        index_to_reduce = [self.variables.index(var) for var in var_to_reduce]

        K_i_i = self.K[np.ix_(index_to_keep, index_to_keep)]
        K_i_j = self.K[np.ix_(index_to_keep, index_to_reduce)]
        K_j_j = self.K[np.ix_(index_to_reduce, index_to_reduce)]
        h_i = self.h[index_to_keep]
        h_j = self.h[index_to_reduce]

        # The values for the reduced variables.
        y = np.array([value for var, value in values]).reshape(len(index_to_reduce), 1)

        phi.variables = [self.variables[index] for index in index_to_keep]
        phi.K = K_i_i
        phi.h = h_i - np.dot(K_i_j, y)
        phi.g = self.g + (np.dot(h_j.T, y) - (0.5 * np.dot(np.dot(y.T, K_j_j), y)))[0][0]

        if not inplace:
            return phi

    def marginalize(self, variables, inplace=True):
        u"""
        Modifies the factor with marginalized values.

        Let C(X,Y ; K, h, g) be some canonical form over X,Y where,

        k = [[K_XX, K_XY],        ;       h = [[h_X],
             [K_YX, K_YY]]                     [h_Y]]

        In this case, the result of the integration operation is a canonical
        from C (K', h', g') given by,

        .. math:: K' = K_{XX} - K_{XY} * {K^{-1}}_{YY} * K_YX
        .. math:: h' = h_X - K_{XY} * {K^{-1}}_{YY} * h_Y
        .. math:: g' = g + 0.5 * (|Y| * log(2*pi) - log(|K_{YY}|) + {h^T}_Y * K_{YY} * h_Y)

        Parameters
        ----------

        variables: list or array-like
                List of variables over which to marginalize.

        inplace: boolean
                If inplace=True it will modify the distribution itself,
                else would return a new distribution.

        Returns
        -------
        CanonicalDistribution or None :
                if inplace=True (default) returns None
                if inplace=False return a new CanonicalDistribution instance

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi = CanonicalDistribution(['X1', 'X2', 'X3'],
        ...                       np.array([[1, -1, 0], [-1, 4, -2], [0, -2, 4]]),
        ...                       np.array([[1], [4], [-1]]), -2)
        >>> phi.K
        array([[ 1, -1,  0],
                [-1,  4, -2],
                [ 0, -2,  4]])

        >>> phi.h
        array([[ 1],
                [ 4],
                [-1]])

        >>> phi.g
        -2

        >>> phi.marginalize(['X3'])

        >>> phi.K
        array([[ 1., -1.],
                [-1.,  3.]])

        >>> phi.h
        array([[ 1. ],
                [ 3.5]])

        >>> phi.g
        0.22579135
        """
        if not isinstance(variables, (list, tuple, np.ndarray)):
            raise TypeError("variables: Expected type list or array-like, "
                            "got type {var_type}".format(var_type=type(variables)))

        if not all([var in self.variables for var in variables]):
            raise ValueError("Variable not in scope.")

        phi = self if inplace else self.copy()

        # index_to_keep -> i vector
        index_to_keep = [self.variables.index(var) for var in self.variables
                         if var not in variables]
        # index_to_marginalize -> j vector
        index_to_marginalize = [self.variables.index(var) for var in variables]

        K_i_i = self.K[np.ix_(index_to_keep, index_to_keep)]
        K_i_j = self.K[np.ix_(index_to_keep, index_to_marginalize)]
        K_j_i = self.K[np.ix_(index_to_marginalize, index_to_keep)]
        K_j_j = self.K[np.ix_(index_to_marginalize, index_to_marginalize)]
        K_j_j_inv = np.linalg.inv(K_j_j)
        h_i = self.h[index_to_keep]
        h_j = self.h[index_to_marginalize]

        phi.variables = [self.variables[index] for index in index_to_keep]

        phi.K = K_i_i - np.dot(np.dot(K_i_j, K_j_j_inv), K_j_i)
        phi.h = h_i - np.dot(np.dot(K_i_j, K_j_j_inv), h_j)
        phi.g = self.g + 0.5 * (len(variables) * np.log(2 * np.pi) -
                                np.log(abs(np.linalg.det(K_j_j))) + np.dot(np.dot(h_j.T, K_j_j), h_j))[0][0]

        if not inplace:
            return phi

    def _operate(self, other, operation, inplace=True):
        """
        Gives the CanonicalDistribution operation (product or divide) with
        the other factor.

        The product of two canonical factors over the same scope
        X is simply:

        C(K1, h1, g1) * C(K2, h2, g2) = C(K1+K2, h1+h2, g1+g2)

        The division of canonical forms is defined analogously:

        C(K1, h1, g1) / C(K2, h2, g2) = C(K1-K2, h1-h2, g1- g2)

        When we have two canonical factors over different scopes X and Y,
        we simply extend the scope of both to make their scopes match and
        then perform the operation of the above equation. The extension of
        the scope is performed by simply adding zero entries to both the K
        matrices and the h vectors.

        Parameters
        ----------
        other: CanonicalFactor
            The CanonicalDistribution to be multiplied.

        operation: String
            'product' for multiplication operation and
            'divide' for division operation.

        Returns
        -------
        CanonicalDistribution or None:
                        if inplace=True (default) returns None
                        if inplace=False returns a new CanonicalDistribution instance.

        Example
        -------
        >>> import numpy as np
        >>> from pgmpy.factors.continuous import CanonicalDistribution
        >>> phi1 = CanonicalDistribution(['x1', 'x2', 'x3'],
                                   np.array([[1, -1, 0], [-1, 4, -2], [0, -2, 4]]),
                                   np.array([[1], [4], [-1]]), -2)
        >>> phi2 = CanonicalDistribution(['x1', 'x2'], np.array([[3, -2], [-2, 4]]),
                                   np.array([[5], [-1]]), 1)

        >>> phi3 = phi1 * phi2
        >>> phi3.K
        array([[ 4., -3.,  0.],
               [-3.,  8., -2.],
               [ 0., -2.,  4.]])
        >>> phi3.h
        array([ 6.,  3., -1.])
        >>> phi3.g
        -1

        >>> phi4 = phi1 / phi2
        >>> phi4.K
        array([[-2.,  1.,  0.],
               [ 1.,  0., -2.],
               [ 0., -2.,  4.]])
        >>> phi4.h
        array([-4.,  5., -1.])
        >>> phi4.g
        -3

        """
        if not isinstance(other, CanonicalDistribution):
            raise TypeError(
                "CanonicalDistribution object can only be multiplied or divided "
                "with an another CanonicalDistribution object. Got {other_type}, "
                "expected CanonicalDistribution.".format(other_type=type(other)))

        phi = self if inplace else self.copy()

        all_vars = self.variables + [var for var in other.variables if var not in self.variables]
        no_of_var = len(all_vars)

        self_var_index = [all_vars.index(var) for var in self.variables]
        other_var_index = [all_vars.index(var) for var in other.variables]

        def _extend_K_scope(K, index):
            ext_K = np.zeros([no_of_var, no_of_var])
            ext_K[np.ix_(index, index)] = K
            return ext_K

        def _extend_h_scope(h, index):
            ext_h = np.zeros(no_of_var).reshape(no_of_var, 1)
            ext_h[index] = h
            return ext_h

        phi.variables = all_vars

        if operation == 'product':
            phi.K = _extend_K_scope(self.K, self_var_index) + _extend_K_scope(other.K, other_var_index)
            phi.h = _extend_h_scope(self.h, self_var_index) + _extend_h_scope(other.h, other_var_index)
            phi.g = self.g + other.g

        else:
            phi.K = _extend_K_scope(self.K, self_var_index) - _extend_K_scope(other.K, other_var_index)
            phi.h = _extend_h_scope(self.h, self_var_index) - _extend_h_scope(other.h, other_var_index)
            phi.g = self.g - other.g

        if not inplace:
            return phi

    def product(self, other, inplace=True):
        """
        Returns the product of two gaussian distributions.

        Parameters
        ----------
        other: CanonicalFactor
            The GaussianDistribution to be multiplied.

        inplace: boolean
            If True, modifies the distribution itself, otherwise returns a new
            CanonicalDistribution object.

        Returns
        -------
        CanonicalDistribution or None:
                    if inplace=True (default) returns None.
                    if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis1 = GD(['x1', 'x2', 'x3'], np.array([[1], [-3], [4]]),
        ...            np.array([[4, 2, -2], [2, 5, -5], [-2, -5, 8]]))
        >>> dis2 = GD(['x3', 'x4'], [1, 2], [[2, 3], [5, 6]])
        >>> dis3 = dis1.product(dis2, inplace=False)
        >>> dis3.covariance
        array([[ 3.6,  1. , -0.4, -0.6],
               [ 1. ,  2.5, -1. , -1.5],
               [-0.4, -1. ,  1.6,  2.4],
               [-1. , -2.5,  4. ,  4.5]])
        >>> dis3.mean
        array([[ 1.6],
               [-1.5],
               [ 1.6],
               [ 3.5]])
        """
        return self._operate(other, operation='product', inplace=inplace)

    def divide(self, other, inplace=True):
        """
        Returns the division of two gaussian distributions.

        Parameters
        ----------
        other: GaussianDistribution
            The GaussianDistribution to be divided.

        inplace: boolean
            If True, modifies the distribution itself, otherwise returns a new
            GaussianDistribution object.

        Returns
        -------
        CanonicalDistribution or None:
                    if inplace=True (default) returns None.
                    if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis1 = GD(['x1', 'x2', 'x3'], np.array([[1], [-3], [4]]),
        ...            np.array([[4, 2, -2], [2, 5, -5], [-2, -5, 8]]))
        >>> dis2 = GD(['x3', 'x4'], [1, 2], [[2, 3], [5, 6]])
        >>> dis3 = dis1.divide(dis2, inplace=False)
        >>> dis3.covariance
        array([[ 3.6,  1. , -0.4, -0.6],
               [ 1. ,  2.5, -1. , -1.5],
               [-0.4, -1. ,  1.6,  2.4],
               [-1. , -2.5,  4. ,  4.5]])
        >>> dis3.mean
        array([[ 1.6],
               [-1.5],
               [ 1.6],
               [ 3.5]])
        """
        return self._operate(other, operation='divide', inplace=inplace)

    def __mul__(self, other):
        return self.product(other, inplace=False)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.divide(other, inplace=False)

    __div__ = __truediv__




import unittest
import numpy as np

from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.models import JunctionTree
from pgmpy.tests import help_functions as hf


class TestJunctionTreeCreation(unittest.TestCase):
    def setUp(self):
        self.graph = JunctionTree()

    def test_add_single_node(self):
        self.graph.add_node(('a', 'b'))
        self.assertListEqual(self.graph.nodes(), [('a', 'b')])

    def test_add_single_node_raises_error(self):
        self.assertRaises(TypeError, self.graph.add_node, 'a')

    def test_add_multiple_nodes(self):
        self.graph.add_nodes_from([('a', 'b'), ('b', 'c')])
        self.assertListEqual(hf.recursive_sorted(self.graph.nodes()),
                             [['a', 'b'], ['b', 'c']])

    def test_add_single_edge(self):
        self.graph.add_edge(('a', 'b'), ('b', 'c'))
        self.assertListEqual(hf.recursive_sorted(self.graph.nodes()),
                             [['a', 'b'], ['b', 'c']])
        self.assertListEqual(sorted([node for edge in self.graph.edges()
                                     for node in edge]),
                             [('a', 'b'), ('b', 'c')])

    def test_add_single_edge_raises_error(self):
        self.assertRaises(ValueError, self.graph.add_edge,
                          ('a', 'b'), ('c', 'd'))

    def test_add_cyclic_path_raises_error(self):
        self.graph.add_edge(('a', 'b'), ('b', 'c'))
        self.graph.add_edge(('b', 'c'), ('c', 'd'))
        self.assertRaises(ValueError, self.graph.add_edge, ('c', 'd'), ('a', 'b'))

    def tearDown(self):
        del self.graph


class TestJunctionTreeMethods(unittest.TestCase):
    def setUp(self):
        self.factor1 = DiscreteFactor(['a', 'b'], [2, 2], np.random.rand(4))
        self.factor2 = DiscreteFactor(['b', 'c'], [2, 2], np.random.rand(4))
        self.factor3 = DiscreteFactor(['d', 'e'], [2, 2], np.random.rand(4))
        self.factor4 = DiscreteFactor(['e', 'f'], [2, 2], np.random.rand(4))
        self.factor5 = DiscreteFactor(['a', 'b', 'e'], [2, 2, 2], np.random.rand(8))

        self.graph1 = JunctionTree()
        self.graph1.add_edge(('a', 'b'), ('b', 'c'))
        self.graph1.add_factors(self.factor1, self.factor2)

        self.graph2 = JunctionTree()
        self.graph2.add_nodes_from([('a', 'b'), ('b', 'c'), ('d', 'e')])
        self.graph2.add_edge(('a', 'b'), ('b', 'c'))
        self.graph2.add_factors(self.factor1, self.factor2, self.factor3)

        self.graph3 = JunctionTree()
        self.graph3.add_edges_from([(('a', 'b'), ('b', 'c')), (('d', 'e'), ('e', 'f'))])
        self.graph3.add_factors(self.factor1, self.factor2, self.factor3, self.factor4)

        self.graph4 = JunctionTree()
        self.graph4.add_edges_from([(('a', 'b', 'e'), ('b', 'c')), (('a', 'b', 'e'), ('e', 'f')),
                                    (('d', 'e'), ('e', 'f'))])
        self.graph4.add_factors(self.factor5, self.factor2, self.factor3, self.factor4)

    def test_check_model(self):
        self.assertRaises(ValueError, self.graph2.check_model)
        self.assertRaises(ValueError, self.graph3.check_model)
        self.assertTrue(self.graph1.check_model())
        self.assertTrue(self.graph4.check_model())

    def tearDown(self):
        del self.factor1
        del self.factor2
        del self.factor3
        del self.factor4
        del self.factor5

        del self.graph1
        del self.graph2
        del self.graph3
        del self.graph4


class TestJunctionTreeCopy(unittest.TestCase):
    def setUp(self):
        self.graph = JunctionTree()

    def test_copy_with_nodes(self):
        self.graph.add_nodes_from([('a', 'b', 'c'), ('a', 'b'), ('a', 'c')])
        self.graph.add_edges_from([(('a', 'b', 'c'), ('a', 'b')),
                                   (('a', 'b', 'c'), ('a', 'c'))])
        graph_copy = self.graph.copy()

        self.graph.remove_edge(('a', 'b', 'c'), ('a', 'c'))
        self.assertFalse(self.graph.has_edge(('a', 'b', 'c'), ('a', 'c')))
        self.assertTrue(graph_copy.has_edge(('a', 'b', 'c'), ('a', 'c')))

        self.graph.remove_node(('a', 'c'))
        self.assertFalse(self.graph.has_node(('a', 'c')))
        self.assertTrue(graph_copy.has_node(('a', 'c')))

        self.graph.add_node(('c', 'd'))
        self.assertTrue(self.graph.has_node(('c', 'd')))
        self.assertFalse(graph_copy.has_node(('c', 'd')))

    def test_copy_with_factors(self):
        self.graph.add_edges_from([[('a', 'b'), ('b', 'c')]])
        phi1 = DiscreteFactor(['a', 'b'], [2, 2], np.random.rand(4))
        phi2 = DiscreteFactor(['b', 'c'], [2, 2], np.random.rand(4))
        self.graph.add_factors(phi1, phi2)
        graph_copy = self.graph.copy()

        self.assertIsInstance(graph_copy, JunctionTree)
        self.assertIsNot(self.graph, graph_copy)
        self.assertEqual(hf.recursive_sorted(self.graph.nodes()),
                         hf.recursive_sorted(graph_copy.nodes()))
        self.assertEqual(hf.recursive_sorted(self.graph.edges()),
                         hf.recursive_sorted(graph_copy.edges()))
        self.assertTrue(graph_copy.check_model())
        self.assertEqual(self.graph.get_factors(), graph_copy.get_factors())

        self.graph.remove_factors(phi1, phi2)
        self.assertTrue(phi1 not in self.graph.factors and phi2 not in self.graph.factors)
        self.assertTrue(phi1 in graph_copy.factors and phi2 in graph_copy.factors)

        self.graph.add_factors(phi1, phi2)
        self.graph.factors[0] = DiscreteFactor(['a', 'b'], [2, 2], np.random.rand(4))
        self.assertNotEqual(self.graph.get_factors()[0], graph_copy.get_factors()[0])
        self.assertNotEqual(self.graph.factors, graph_copy.factors)

    def test_copy_with_factorchanges(self):
        self.graph.add_edges_from([[('a', 'b'), ('b', 'c')]])
        phi1 = DiscreteFactor(['a', 'b'], [2, 2], np.random.rand(4))
        phi2 = DiscreteFactor(['b', 'c'], [2, 2], np.random.rand(4))
        self.graph.add_factors(phi1, phi2)
        graph_copy = self.graph.copy()

        self.graph.factors[0].reduce([('a', 0)])
        self.assertNotEqual(self.graph.factors[0].scope(), graph_copy.factors[0].scope())
        self.assertNotEqual(self.graph, graph_copy)
        self.graph.factors[1].marginalize(['b'])
        self.assertNotEqual(self.graph.factors[1].scope(), graph_copy.factors[1].scope())
        self.assertNotEqual(self.graph, graph_copy)

    def tearDown(self):
        del self.graph



import itertools
from operator import mul

import numpy as np

from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.independencies import Independencies
from pgmpy.extern.six.moves import range, zip
from pgmpy.extern import six


class JointProbabilityDistribution(DiscreteFactor):
    """
    Base class for Joint Probability Distribution

    Public Methods
    --------------
    conditional_distribution(values)
    create_bayesian_model()
    get_independencies()
    pmap()
    marginal_distribution(variables)
    minimal_imap()
    is_imap(model)
    """

    def __init__(self, variables, cardinality, values):
        """
        Initialize a Joint Probability Distribution class.

        Defined above, we have the following mapping from variable
        assignments to the index of the row vector in the value field:

        +-----+-----+-----+-------------------------+
        |  x1 |  x2 |  x3 |    P(x1, x2, x2)        |
        +-----+-----+-----+-------------------------+
        | x1_0| x2_0| x3_0|    P(x1_0, x2_0, x3_0)  |
        +-----+-----+-----+-------------------------+
        | x1_1| x2_0| x3_0|    P(x1_1, x2_0, x3_0)  |
        +-----+-----+-----+-------------------------+
        | x1_0| x2_1| x3_0|    P(x1_0, x2_1, x3_0)  |
        +-----+-----+-----+-------------------------+
        | x1_1| x2_1| x3_0|    P(x1_1, x2_1, x3_0)  |
        +-----+-----+-----+-------------------------+
        | x1_0| x2_0| x3_1|    P(x1_0, x2_0, x3_1)  |
        +-----+-----+-----+-------------------------+
        | x1_1| x2_0| x3_1|    P(x1_1, x2_0, x3_1)  |
        +-----+-----+-----+-------------------------+
        | x1_0| x2_1| x3_1|    P(x1_0, x2_1, x3_1)  |
        +-----+-----+-----+-------------------------+
        | x1_1| x2_1| x3_1|    P(x1_1, x2_1, x3_1)  |
        +-----+-----+-----+-------------------------+

        Parameters
        ----------
        variables: list
            List of scope of Joint Probability Distribution.
        cardinality: list, array_like
            List of cardinality of each variable
        value: list, array_like
            List or array of values of factor.
            A Joint Probability Distribution's values are stored in a row
            vector in the value using an ordering such that the left-most
            variables as defined in the variable field cycle through their
            values the fastest.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 2, 2], np.ones(8)/8)
        >>> print(prob)
        x1    x2    x3      P(x1,x2,x3)
        ----  ----  ----  -------------
        x1_0  x2_0  x3_0         0.1250
        x1_0  x2_0  x3_1         0.1250
        x1_0  x2_1  x3_0         0.1250
        x1_0  x2_1  x3_1         0.1250
        x1_1  x2_0  x3_0         0.1250
        x1_1  x2_0  x3_1         0.1250
        x1_1  x2_1  x3_0         0.1250
        x1_1  x2_1  x3_1         0.1250
       """
        if np.isclose(np.sum(values), 1):
            super(JointProbabilityDistribution, self).__init__(variables, cardinality, values)
        else:
            raise ValueError("The probability values doesn't sum to 1.")

    def __repr__(self):
        var_card = ", ".join(['{var}:{card}'.format(var=var, card=card)
                              for var, card in zip(self.variables, self.cardinality)])
        return "<Joint Distribution representing P({var_card}) at {address}>".format(address=hex(id(self)),
                                                                                     var_card=var_card)

    def __str__(self):
        if six.PY2:
            return self._str(phi_or_p='P', tablefmt='pqsl')
        else:
            return self._str(phi_or_p='P')

    def marginal_distribution(self, variables, inplace=True):
        """
        Returns the marginal distribution over variables.

        Parameters
        ----------
        variables: string, list, tuple, set, dict
                Variable or list of variables over which marginal distribution needs
                to be calculated
        inplace: Boolean (default True)
                If False return a new instance of JointProbabilityDistribution

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> values = np.random.rand(12)
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 3, 2], values/np.sum(values))
        >>> prob.marginal_distribution(['x1', 'x2'])
        >>> print(prob)
        x1    x2      P(x1,x2)
        ----  ----  ----------
        x1_0  x2_0      0.1502
        x1_0  x2_1      0.1626
        x1_0  x2_2      0.1197
        x1_1  x2_0      0.2339
        x1_1  x2_1      0.1996
        x1_1  x2_2      0.1340
        """
        return self.marginalize(list(set(list(self.variables)) -
                                     set(variables if isinstance(
                                         variables, (list, set, dict, tuple)) else [variables])),
                                inplace=inplace)

    def check_independence(self, event1, event2, event3=None, condition_random_variable=False):
        """
        Check if the Joint Probability Distribution satisfies the given independence condition.

        Parameters
        ----------
        event1: list
            random variable whose independence is to be checked.
        event2: list
            random variable from which event1 is independent.
        values: 2D array or list like or 1D array or list like
            A 2D list of tuples of the form (variable_name, variable_state).
            A 1D list or array-like to condition over randome variables (condition_random_variable must be True)
            The values on which to condition the Joint Probability Distribution.
        condition_random_variable: Boolean (Default false)
            If true and event3 is not None than will check independence condition over random variable.

        For random variables say X, Y, Z to check if X is independent of Y given Z.
        event1 should be either X or Y.
        event2 should be either Y or X.
        event3 should Z.

        Examples
        --------
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution as JPD
        >>> prob = JPD(['I','D','G'],[2,2,3],
                       [0.126,0.168,0.126,0.009,0.045,0.126,0.252,0.0224,0.0056,0.06,0.036,0.024])
        >>> prob.check_independence(['I'], ['D'])
        True
        >>> prob.check_independence(['I'], ['D'], [('G', 1)])  # Conditioning over G_1
        False
        >>> # Conditioning over random variable G
        >>> prob.check_independence(['I'], ['D'], ('G',), condition_random_variable=True)
        False
        """
        JPD = self.copy()
        if isinstance(event1, six.string_types):
            raise TypeError('Event 1 should be a list or array-like structure')

        if isinstance(event2, six.string_types):
            raise TypeError('Event 2 should be a list or array-like structure')

        if event3:
            if isinstance(event3, six.string_types):
                raise TypeError('Event 3 cannot of type string')

            elif condition_random_variable:
                if not all(isinstance(var, six.string_types) for var in event3):
                    raise TypeError('Event3 should be a 1d list of strings')
                event3 = list(event3)
                # Using the definition of conditional independence
                # If P(X,Y|Z) = P(X|Z)*P(Y|Z)
                # This can be expanded to P(X,Y,Z)*P(Z) == P(X,Z)*P(Y,Z)
                phi_z = JPD.marginal_distribution(event3, inplace=False).to_factor()
                for variable_pair in itertools.product(event1, event2):
                    phi_xyz = JPD.marginal_distribution(event3 + list(variable_pair), inplace=False).to_factor()
                    phi_xz = JPD.marginal_distribution(event3 + [variable_pair[0]], inplace=False).to_factor()
                    phi_yz = JPD.marginal_distribution(event3 + [variable_pair[1]], inplace=False).to_factor()
                    if phi_xyz * phi_z != phi_xz * phi_yz:
                        return False
                return True
            else:
                JPD.conditional_distribution(event3)

        for variable_pair in itertools.product(event1, event2):
            if (JPD.marginal_distribution(variable_pair, inplace=False) !=
                    JPD.marginal_distribution(variable_pair[0], inplace=False) *
                    JPD.marginal_distribution(variable_pair[1], inplace=False)):
                return False
        return True

    def get_independencies(self, condition=None):
        """
        Returns the independent variables in the joint probability distribution.
        Returns marginally independent variables if condition=None.
        Returns conditionally independent variables if condition!=None

        Parameter
        ---------
        condition: array_like
                Random Variable on which to condition the Joint Probability Distribution.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 3, 2], np.ones(12)/12)
        >>> prob.get_independencies()
        (x1 _|_ x2)
        (x1 _|_ x3)
        (x2 _|_ x3)
        """
        JPD = self.copy()
        if condition:
            JPD.conditional_distribution(condition)
        independencies = Independencies()
        for variable_pair in itertools.combinations(list(JPD.variables), 2):
            if (JPD.marginal_distribution(variable_pair, inplace=False) ==
                    JPD.marginal_distribution(variable_pair[0], inplace=False) *
                    JPD.marginal_distribution(variable_pair[1], inplace=False)):
                independencies.add_assertions(variable_pair)
        return independencies

    def conditional_distribution(self, values, inplace=True):
        """
        Returns Conditional Probability Distribution after setting values to 1.

        Parameters
        ----------
        values: list or array_like
            A list of tuples of the form (variable_name, variable_state).
            The values on which to condition the Joint Probability Distribution.
        inplace: Boolean (default True)
            If False returns a new instance of JointProbabilityDistribution

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 2, 2], np.ones(8)/8)
        >>> prob.conditional_distribution([('x1', 1)])
        >>> print(prob)
        x2    x3      P(x2,x3)
        ----  ----  ----------
        x2_0  x3_0      0.2500
        x2_0  x3_1      0.2500
        x2_1  x3_0      0.2500
        x2_1  x3_1      0.2500
        """
        JPD = self if inplace else self.copy()
        JPD.reduce(values)
        JPD.normalize()
        if not inplace:
            return JPD

    def copy(self):
        """
        Returns A copy of JointProbabilityDistribution object

        Examples
        ---------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 3, 2], np.ones(12)/12)
        >>> prob_copy = prob.copy()
        >>> prob_copy.values == prob.values
        True
        >>> prob_copy.variables == prob.variables
        True
        >>> prob_copy.variables[1] = 'y'
        >>> prob_copy.variables == prob.variables
        False
        """
        return JointProbabilityDistribution(self.scope(), self.cardinality, self.values)

    def minimal_imap(self, order):
        """
        Returns a Bayesian Model which is minimal IMap of the Joint Probability Distribution
        considering the order of the variables.

        Parameters
        ----------
        order: array-like
            The order of the random variables.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 3, 2], np.ones(12)/12)
        >>> bayesian_model = prob.minimal_imap(order=['x2', 'x1', 'x3'])
        >>> bayesian_model
        <pgmpy.models.models.models at 0x7fd7440a9320>
        >>> bayesian_model.edges()
        [('x1', 'x3'), ('x2', 'x3')]
        """
        from pgmpy.models import BayesianModel

        def get_subsets(u):
            for r in range(len(u) + 1):
                for i in itertools.combinations(u, r):
                    yield i

        G = BayesianModel()
        for variable_index in range(len(order)):
            u = order[:variable_index]
            for subset in get_subsets(u):
                if (len(subset) < len(u) and
                        self.check_independence([order[variable_index]], set(u) - set(subset), subset, True)):
                    G.add_edges_from([(variable, order[variable_index]) for variable in subset])
        return G

    def is_imap(self, model):
        """
        Checks whether the given BayesianModel is Imap of JointProbabilityDistribution

        Parameters
        -----------
        model : An instance of BayesianModel Class, for which you want to
            check the Imap

        Returns
        --------
        boolean : True if given bayesian model is Imap for Joint Probability Distribution
                False otherwise
        Examples
        --------
        >>> from pgmpy.models import BayesianModel
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> bm = BayesianModel([('diff', 'grade'), ('intel', 'grade')])
        >>> diff_cpd = TabularCPD('diff', 2, [[0.2], [0.8]])
        >>> intel_cpd = TabularCPD('intel', 3, [[0.5], [0.3], [0.2]])
        >>> grade_cpd = TabularCPD('grade', 3,
        ...                        [[0.1,0.1,0.1,0.1,0.1,0.1],
        ...                         [0.1,0.1,0.1,0.1,0.1,0.1],
        ...                         [0.8,0.8,0.8,0.8,0.8,0.8]],
        ...                        evidence=['diff', 'intel'],
        ...                        evidence_card=[2, 3])
        >>> bm.add_cpds(diff_cpd, intel_cpd, grade_cpd)
        >>> val = [0.01, 0.01, 0.08, 0.006, 0.006, 0.048, 0.004, 0.004, 0.032,
                   0.04, 0.04, 0.32, 0.024, 0.024, 0.192, 0.016, 0.016, 0.128]
        >>> JPD = JointProbabilityDistribution(['diff', 'intel', 'grade'], [2, 3, 3], val)
        >>> JPD.is_imap(bm)
        True
        """
        from pgmpy.models import BayesianModel
        if not isinstance(model, BayesianModel):
            raise TypeError("model must be an instance of BayesianModel")
        factors = [cpd.to_factor() for cpd in model.get_cpds()]
        factor_prod = six.moves.reduce(mul, factors)
        JPD_fact = DiscreteFactor(self.variables, self.cardinality, self.values)
        if JPD_fact == factor_prod:
            return True
        else:
            return False

    def to_factor(self):
        """
        Returns JointProbabilityDistribution as a DiscreteFactor object

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.discrete import JointProbabilityDistribution
        >>> prob = JointProbabilityDistribution(['x1', 'x2', 'x3'], [2, 3, 2], np.ones(12)/12)
        >>> phi = prob.to_factor()
        >>> type(phi)
        pgmpy.factors.DiscreteFactor.DiscreteFactor
        """
        return DiscreteFactor(self.variables, self.cardinality, self.values)

    def pmap(self):
        pass



from .base import BaseDistribution
from .CustomDistribution import CustomDistribution
from .GaussianDistribution import GaussianDistribution


__all__ = ['BaseDistribution',
           'CustomDistribution',
           'GaussianDistribution'
           ]



import unittest

import pandas as pd

from pgmpy.models import BayesianModel
from pgmpy.estimators import BicScore


class TestBicScore(unittest.TestCase):
    def setUp(self):
        self.d1 = pd.DataFrame(data={'A': [0, 0, 1], 'B': [0, 1, 0], 'C': [1, 1, 0], 'D': ['X', 'Y', 'Z']})
        self.m1 = BayesianModel([('A', 'C'), ('B', 'C'), ('D', 'B')])
        self.m2 = BayesianModel([('C', 'A'), ('C', 'B'), ('A', 'D')])

        # data_link - "https://www.kaggle.com/c/titanic/download/train.csv"
        self.titanic_data = pd.read_csv('pgmpy/tests/test_estimators/testdata/titanic_train.csv')
        self.titanic_data2 = self.titanic_data[["Survived", "Sex", "Pclass"]]

    def test_score(self):
        self.assertAlmostEqual(BicScore(self.d1).score(self.m1), -10.698440814229318)
        self.assertEqual(BicScore(self.d1).score(BayesianModel()), 0)

    def test_score_titanic(self):
        scorer = BicScore(self.titanic_data2)
        titanic = BayesianModel([("Sex", "Survived"), ("Pclass", "Survived")])
        self.assertAlmostEqual(scorer.score(titanic), -1896.7250012840179)
        titanic2 = BayesianModel([("Pclass", "Sex"), ])
        titanic2.add_nodes_from(["Sex", "Survived", "Pclass"])
        self.assertLess(scorer.score(titanic2), scorer.score(titanic))

    def tearDown(self):
        del self.d1
        del self.m1
        del self.m2
        del self.titanic_data
        del self.titanic_data2



import unittest

import numpy as np
import numpy.testing as np_test

from pgmpy.readwrite import BIFReader, BIFWriter
from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD
from pgmpy.extern.six.moves import map, range


class TestBIFReader(unittest.TestCase):

    def setUp(self):

        self.reader = BIFReader(string="""
// Bayesian Network in the Interchange Format
// Produced by BayesianNetworks package in JavaBayes
// Output created Sun Nov 02 17:49:49 GMT+00:00 1997
// Bayesian network
network "Dog-Problem" { //5 variables and 5 probability distributions
        property "credal-set constant-density-bounded 1.1" ;
}
variable  "light-on" { //2 values
        type discrete[2] {  "true"  "false" };
        property "position = (218, 195)" ;
}
variable  "bowel-problem" { //2 values
        type discrete[2] {  "true"  "false" };
        property "position = (335, 99)" ;
}
variable  "dog-out" { //2 values
        type discrete[2] {  "true"  "false" };
        property "position = (300, 195)" ;
}
variable  "hear-bark" { //2 values
        type discrete[2] {  "true"  "false" };
        property "position = (296, 268)" ;
}
variable  "family-out" { //2 values
        type discrete[2] {  "true"  "false" };
        property "position = (257, 99)" ;
}
probability (  "light-on"  "family-out" ) { //2 variable(s) and 4 values
        (true) 0.6 0.4 ;
        (false) 0.05 0.95 ;
}
probability (  "bowel-problem" ) { //1 variable(s) and 2 values
        table 0.01 0.99 ;
}
probability (  "dog-out"  "bowel-problem"  "family-out" ) { //3 variable(s) and 8 values
        table 0.99 0.97 0.9 0.3 0.01 0.03 0.1 0.7 ;
}
probability (  "hear-bark"  "dog-out" ) { //2 variable(s) and 4 values
        table 0.7 0.01 0.3 0.99 ;
}
probability (  "family-out" ) { //1 variable(s) and 2 values
        table 0.15 0.85 ;
}
""")

    def test_network_name(self):

        name_expected = 'Dog-Problem'
        self.assertEqual(self.reader.network_name, name_expected)

    def test_get_variables(self):

        var_expected = ['light-on', 'bowel-problem', 'dog-out',
                        'hear-bark', 'family-out']
        self.assertListEqual(self.reader.get_variables(), var_expected)

    def test_states(self):

        states_expected = {'bowel-problem': ['true', 'false'],
                           'dog-out': ['true', 'false'],
                           'family-out': ['true', 'false'],
                           'hear-bark': ['true', 'false'],
                           'light-on': ['true', 'false']}
        states = self.reader.get_states()
        for variable in states_expected:
            self.assertListEqual(states_expected[variable], states[variable])

    def test_get_property(self):

        property_expected = {'bowel-problem': ['position = (335, 99)'],
                             'dog-out': ['position = (300, 195)'],
                             'family-out': ['position = (257, 99)'],
                             'hear-bark': ['position = (296, 268)'],
                             'light-on': ['position = (218, 195)']}
        prop = self.reader.get_property()
        for variable in property_expected:
            self.assertListEqual(property_expected[variable],
                                 prop[variable])

    def test_get_values(self):

        cpd_expected = {'bowel-problem': np.array([[0.01],
                                                   [0.99]]),
                        'dog-out': np.array([[0.99, 0.97, 0.9, 0.3],
                                             [0.01, 0.03, 0.1, 0.7]]),
                        'family-out': np.array([[0.15],
                                                [0.85]]),
                        'hear-bark': np.array([[0.7, 0.01],
                                               [0.3, 0.99]]),
                        'light-on': np.array([[0.6, 0.05],
                                              [0.4, 0.95]])}
        cpd = self.reader.variable_cpds
        for variable in cpd_expected:
            np_test.assert_array_equal(cpd_expected[variable],
                                       cpd[variable])

    def test_get_parents(self):

        parents_expected = {'bowel-problem': [],
                            'dog-out': ['bowel-problem', 'family-out'],
                            'family-out': [],
                            'hear-bark': ['dog-out'],
                            'light-on': ['family-out']}
        parents = self.reader.get_parents()
        for variable in parents_expected:
            self.assertListEqual(parents_expected[variable],
                                 parents[variable])

    def test_get_edges(self):

        edges_expected = [['family-out', 'dog-out'],
                          ['bowel-problem', 'dog-out'],
                          ['family-out', 'light-on'],
                          ['dog-out', 'hear-bark']]
        self.assertListEqual(sorted(self.reader.variable_edges),
                             sorted(edges_expected))

    def test_get_model(self):
        edges_expected = [('family-out', 'dog-out'),
                          ('bowel-problem', 'dog-out'),
                          ('family-out', 'light-on'),
                          ('dog-out', 'hear-bark')]
        nodes_expected = ['bowel-problem', 'hear-bark', 'light-on',
                          'dog-out', 'family-out']
        edge_expected = {'bowel-problem': {'dog-out': {'weight': None}},
                         'dog-out': {'hear-bark': {'weight': None}},
                         'family-out': {'dog-out': {'weight': None},
                                        'light-on': {'weight': None}},
                         'hear-bark': {},
                         'light-on': {}}
        node_expected = {'bowel-problem': {'weight': None,
                                           'position': '(335, 99)'},
                         'dog-out': {'weight': None,
                                     'position': '(300, 195)'},
                         'family-out': {'weight': None,
                                        'position': '(257, 99)'},
                         'hear-bark': {'weight': None,
                                       'position': '(296, 268)'},
                         'light-on': {'weight': None,
                                      'position': '(218, 195)'}}
        cpds_expected = [np.array([[0.01],
                                   [0.99]]),
                         np.array([[0.99, 0.97, 0.9, 0.3],
                                   [0.01, 0.03, 0.1, 0.7]]),
                         np.array([[0.15],
                                   [0.85]]),
                         np.array([[0.7, 0.01],
                                   [0.3, 0.99]]),
                         np.array([[0.6, 0.05],
                                   [0.4, 0.95]])]
        model = self.reader.get_model()
        for cpd_index in range(0, len(cpds_expected)):
            np_test.assert_array_equal(model.get_cpds()[cpd_index].get_values(),
                                       cpds_expected[cpd_index])
        self.assertDictEqual(model.node, node_expected)
        self.assertDictEqual(model.edge, edge_expected)
        self.assertListEqual(sorted(model.nodes()), sorted(nodes_expected))
        self.assertListEqual(sorted(model.edges()), sorted(edges_expected))

    def tearDown(self):
        del self.reader


class TestBIFWriter(unittest.TestCase):

    def setUp(self):
        edges = [['family-out', 'dog-out'],
                 ['bowel-problem', 'dog-out'],
                 ['family-out', 'light-on'],
                 ['dog-out', 'hear-bark']]

        cpds = {'bowel-problem': np.array([[0.01],
                                           [0.99]]),
                'dog-out': np.array([[0.99, 0.01, 0.97, 0.03],
                                     [0.9, 0.1, 0.3, 0.7]]),
                'family-out': np.array([[0.15],
                                        [0.85]]),
                'hear-bark': np.array([[0.7, 0.3],
                                       [0.01, 0.99]]),
                'light-on': np.array([[0.6, 0.4],
                                      [0.05, 0.95]])}

        states = {'bowel-problem': ['true', 'false'],
                  'dog-out': ['true', 'false'],
                  'family-out': ['true', 'false'],
                  'hear-bark': ['true', 'false'],
                  'light-on': ['true', 'false']}

        parents = {'bowel-problem': [],
                   'dog-out': ['family-out', 'bowel-problem'],
                   'family-out': [],
                   'hear-bark': ['dog-out'],
                   'light-on': ['family-out']}

        properties = {'bowel-problem': ['position = (335, 99)'],
                      'dog-out': ['position = (300, 195)'],
                      'family-out': ['position = (257, 99)'],
                      'hear-bark': ['position = (296, 268)'],
                      'light-on': ['position = (218, 195)']}

        self.model = BayesianModel(edges)

        tabular_cpds = []
        for var in sorted(cpds.keys()):
            values = cpds[var]
            cpd = TabularCPD(var, len(states[var]), values,
                             evidence=parents[var],
                             evidence_card=[len(states[evidence_var])
                                            for evidence_var in parents[var]])
            tabular_cpds.append(cpd)
        self.model.add_cpds(*tabular_cpds)

        for node, properties in properties.items():
            for prop in properties:
                prop_name, prop_value = map(lambda t: t.strip(), prop.split('='))
                self.model.node[node][prop_name] = prop_value

        self.writer = BIFWriter(model=self.model)

    def test_str(self):
        self.expected_string = """network unknown {
}
variable bowel-problem {
    type discrete [ 2 ] { bowel-problem_0, bowel-problem_1 };
    property position = (335, 99) ;
}
variable dog-out {
    type discrete [ 2 ] { dog-out_0, dog-out_1 };
    property position = (300, 195) ;
}
variable family-out {
    type discrete [ 2 ] { family-out_0, family-out_1 };
    property position = (257, 99) ;
}
variable hear-bark {
    type discrete [ 2 ] { hear-bark_0, hear-bark_1 };
    property position = (296, 268) ;
}
variable light-on {
    type discrete [ 2 ] { light-on_0, light-on_1 };
    property position = (218, 195) ;
}
probability ( bowel-problem ) {
    table 0.01, 0.99 ;
}
probability ( dog-out | bowel-problem, family-out ) {
    table 0.99, 0.01, 0.97, 0.03, 0.9, 0.1, 0.3, 0.7 ;
}
probability ( family-out ) {
    table 0.15, 0.85 ;
}
probability ( hear-bark | dog-out ) {
    table 0.7, 0.3, 0.01, 0.99 ;
}
probability ( light-on | family-out ) {
    table 0.6, 0.4, 0.05, 0.95 ;
}
"""
        self.maxDiff = None
        self.assertEqual(self.writer.__str__(), self.expected_string)
